""" Tests for parser utils. """
import pytest

from excitingtools.parser_utils.parser_utils import json_convert, standardise_fortran_exponent, \
    convert_single_entry, convert_string_dict


@pytest.mark.parametrize(
    "test_input,expected",
    [("true", True), ("false", False), ("fromscratch", "fromscratch"), ("skip", "skip"), ("3", 3), ("-1", -1),
     ("2.34", 2.34), ("5.1e3", 5100)]
)
def test_convert_json(test_input, expected):
    assert json_convert(test_input) == expected


@pytest.mark.parametrize("test_input,expected", [("23.2D-1", 2.32), ("1q0", 1)])
def test_convert_fortran_exponent(test_input, expected):
    assert standardise_fortran_exponent(test_input, return_as_str=False) == expected


@pytest.mark.parametrize(
    "test_input,expected",
    [("true", True), ("skip", "skip"), ("3", 3), ("2.34", 2.34), ("true false false", [True, False, False]),
     ("1.3 2.3e4 3 90", [1.3, 2.3e4, 3, 90]), ("4 4 3", [4, 4, 3]), ("1 2", [1, 2]), ("ab cd", "ab cd")]
)
def test_convert_single_entry(test_input, expected):
    assert convert_single_entry(test_input) == expected


def test_convert_single_entry_to_int():
    converted_int = convert_single_entry("3")
    assert converted_int == 3
    # need to test really the type to detect the error
    assert isinstance(converted_int, int)


def test_convert_string_dict():
    string_dict = {"a": "1", "b": "b", "c": "false", "d": "3 3 3"}
    ref_dict = {"a": 1, "b": "b", "c": False, "d": [3, 3, 3]}
    assert convert_string_dict(string_dict) == ref_dict



""" Dictionary Utilities.
"""
import copy
import json
import sys
from collections.abc import Mapping, Hashable, KeysView
from typing import Iterator, Union

import numpy as np


def common_iterable(obj: Union[dict, list]):
    """
    Create an iterable dict or tuple, such that one can iterate over a
    dictionary or list with the same syntax:

        for index in common_iterable(d):
            print('index = key or integer index:', index)
            print('d[index] = dictionary value or list element:', d[index])
    """
    if isinstance(obj, dict):
        return obj
    else:
        return (index for index, value in enumerate(obj))


def __container_converter(data: Union[list, dict]) -> dict:
    """
    Converts string versions of numerical data from input dict into numerical data.

    :param dict data: Dictionary with string values.
    :return:dict new_data: Dictionary with all string literal values converted to numerical values.
    """
    np_convert = {np.float64: float, np.int32: int}

    for element in common_iterable(data):
        if isinstance(data[element], dict) or isinstance(data[element], list):
            data[element] = __container_converter(data[element])
        elif isinstance(data[element], np.ndarray):
            data[element] = data[element].tolist()
        elif isinstance(data[element], (np.float64, np.int32)):
            value = data[element]
            data[element] = np_convert[type(value)](value)
        elif isinstance(data[element], Hashable):
            try:
                data[element] = json.loads(data[element])
            except:
                pass
        else:
            message = 'Type not converted by dict converter: ' + str(
                type(data[element]))
            sys.exit(message)

    return data


def container_converter(data: dict) -> dict:
    """
    Converts string representations of numerical data into numerical data.

    Acts on string data in the input dict.

    :param dict data: Dictionary to be copied containing string values.
    :return: dict new_data: Copied dictionary with all string literal values converted to numerical values.
    """
    new_data = copy.deepcopy(data)
    return __container_converter(new_data)


def serialise_dict_values(d):
    """
    Mutate dictionary values that are objects, to dictionaries.
    Function works recursively, so no type-hinting.

    If any value within the dictionary is a set, it is converted to a list because
    common_iterable is not compatible with sets (sets cannot be indexed).

    If any value within the dictionary is a tuple, it is converted to a list because
    tuples are immutable, and therefore not consistent with an implementation that
    mutates the input.

    Sets and tuples are converted to lists regardless of whether they contain objects.

    :param dict input_dict: Dictionary.
    :param dict output_dict: Dictionary with any object values converted to dictionaries.
    """

    for index in common_iterable(d):
        # Can't iterate over a set with an index, and tuple cannot be mutated
        if isinstance(d[index], (set, tuple)):
            d[index] = list(d[index])

        # Class instances with the __dict__ attribute
        if '__dict__' in dir(d[index]):
            d[index] = vars(d[index])

        # Recursively apply routine to other containers
        elif not isinstance(d[index], Hashable):
            d[index] = serialise_dict_values(d[index])

        # pass over hashable values and np arrays
        else:
            pass

    return d


def get_hashable_entries(nested: dict) -> Iterator[tuple]:
    """
    Create a generator of all hashable values, removing nesting.

    Note, this will fail if a cross edge or a back edge is included in the dict
    i.e. a self-reference of the form:

        A = {key1: value, key2: A}

    However, for the use case of flattening parsed results, this should never be an issue.

    Reference: https://stackoverflow.com/questions/10756427/loop-through-all-nested-dictionary-values

    :param nested: Nested dictionary with a mix and hashable and non-hashable values.
    :return: Iterator[tuple] (key, value): Generator for tuples, containing keys and values.
    """
    for key, value in nested.items():
        if isinstance(value, Mapping):
            yield from get_hashable_entries(value)
        else:
            yield key, value


def delete_nested_key(dictionary: dict, key_chain: list):
    """
    Delete a key at any level of a nested dictionary.

    :param dict dictionary: dictionary, mutated by function.
    :param list key_chain: List of keys leading to the item to be deleted.
    """
    if len(key_chain) == 1:
        del dictionary[key_chain[0]]
        return
    else:
        delete_nested_key(dictionary[key_chain[0]], key_chain[1:])


def check_valid_keys(input_keys: Union[list, set, tuple, KeysView],
                     valid_keys: Union[list, set, tuple, KeysView],
                     name: str = ''):
    """ Check that a given set of input keys are valid.

    :param input_keys: Input keys
    :param valid_keys: Valid keys
    :param name: Optional name for error message
    """
    erroneous_inputs = set(input_keys) - set(valid_keys)
    if erroneous_inputs:
        raise ValueError(f'{name} keys are not valid: {erroneous_inputs}')



""" KS (ground state) Band structure Parser, returning to an object.
"""

import os
import numpy as np
from excitingtools.dataclasses.band_structure import BandData
from excitingtools.exciting_dict_parsers.properties_parser import parse_band_structure_xml, parse_band_structure_dat
from pathlib import Path


def parse_band_structure(file_name: str) -> BandData:
    """ High-level parser for KS band structure. Calls dictionary parsers to parse information from both
    "bandstructure.xml" and "bandstructure.dat" files.

    :param str file_name: File name for band structure such as "bandstructure.dat", "bandstructure.xml" or
                          "bandstructure"
    :return: BandData object.
    """
    name = os.path.splitext(file_name)[0]
    data_from_xml = parse_band_structure_xml(name + '.xml')
    data_from_dat = parse_band_structure_dat(name + '.dat')

    e_fermi = float(np.loadtxt(Path(file_name).parent / "EFERMI.OUT"))

    return BandData(bands=data_from_xml['band_energies'],
                    k_points=data_from_dat['k_points'],
                    e_fermi=e_fermi,
                    flattened_k_points=data_from_xml['k_points_along_band'],
                    vertices=data_from_xml['vertices'])



""" Parse input XML data directly into corresponding python classes.
"""

from excitingtools.exciting_dict_parsers.input_parser import parse_input_xml as parse_input_xml_to_dict
from excitingtools.input.input_xml import ExcitingInputXML
from excitingtools.parser_utils.parser_decorators import xml_root


@xml_root
def parse_input_xml(root) -> ExcitingInputXML:
    """ Parse exciting input.xml into the corresponding python ExcitingInput Objects.
    :param root: Input for the parser.
    :returns: the exciting input object
    """
    element_dict: dict = parse_input_xml_to_dict(root)
    return ExcitingInputXML(**element_dict)



"""Utilities to aid in writing and formatting XML
"""
import re
from xml.dom import minidom
from xml.etree import ElementTree


def xml_tree_to_pretty_str(elem: ElementTree.Element) -> str:
    """Convert an XML element to a pretty string.

    :param ElementTree.Element elem: Element/ element tree
    :return str : XML tree string, with pretty formatting.
    """
    rough_string = ElementTree.tostring(elem, 'utf-8')
    reparsed = minidom.parseString(rough_string)
    return reparsed.toprettyxml(indent="\t")


def line_reformatter(input_str: str) -> str:
    """Identify attributes of an XML string element, and reformat them such that they are on new lines.

    NOTE: This could be split into two routines. One that finds the (start, end) of
    each attribute, and one that reformats the string, given these indices.
    (or refactored completely - current solution is not very elegant).

    Example
    ----------
    Takes:
        input_str = '\t<groundstate do="fromscratch" ngridk="6 6 6" nosource="false" rgkmax="8.0"
                     tforce="true" vkloff="0 0 0" xctype="GGA_PBE_SOL"> </groundstate>'

    Returns:
        reformatted_str =
        '<groundstate
             do="fromscratch"
             ngridk="6 6 6"
             nosource="false"
             rgkmax="8.0"
             tforce="true"
             vkloff="0 0 0"
             xctype="GGA_PBE_SOL">
        </groundstate>'

    There are 3 possibilities: the xml element has further subelements, than it ends only with a single '>',
    if the xml element has only attributes, than there are two options to close the element: either long with a
    '> </tag>' or short with a '/>'.

    :param str input_str: Input string, opened and closed with an XML element.
    :return str reformatted_str: Reformatted form of input_str
    """
    full_tag = input_str.split(" ")[0]
    tag = full_tag.strip()
    number_of_tag_indents = len(full_tag) - len(tag)

    tag_indent = '\t' * number_of_tag_indents
    attr_indent = tag_indent + ' ' * 3

    # Get rid of format characters, like \n, \t etc
    input_str = input_str.strip()

    # Isolate attributes according to position of quotation marks in string
    # (cannot use whitespaces to split)
    quote_indices = [x.start() for x in re.finditer('\"', input_str)]
    closing_quote_indices = quote_indices[1::2]
    attribute_start_indices = [len(tag) + 1] + [i + 1 for i in
                                                closing_quote_indices[:-1]]

    reformatted_str = tag_indent + tag + '\n'

    for i in range(0, len(attribute_start_indices)):
        i1 = attribute_start_indices[i]
        i2 = closing_quote_indices[i]
        attribute_str = input_str[i1:i2 + 1].strip()
        reformatted_str += attr_indent + attribute_str + '\n'

    # short ending option:
    if input_str[-2:] == '/>':
        return reformatted_str[:-1] + '/>'
    reformatted_str = reformatted_str[:-1] + '>'
    # no ending option:
    if not input_str[-(len(tag) + 2):-len(tag)] == '</':
        return reformatted_str
    # long ending option:
    reformatted_str += "\n" + tag_indent + input_str[-(len(tag) + 2):] + ''
    return reformatted_str


def prettify_tag_attributes(xml_string: str) -> str:
    """Prettify XML string formatting of attributes.

    The routine finds the lines containing an XML element, applies
    a line_reformatter, and replaces the line. Only for long lines with more than 2 attributes.

    Example usage:
    ```
        string = <groundstate do="fromscratch" ngridk="6 6 6" nosource="false" rgkmax="8.0" tforce="true" vkloff="0 0 0"
                  xctype="GGA_PBE_SOL"> </groundstate>
        pretty_string = prettify_tag_attributes(string)
        print(pretty_string)
        > <groundstate
             do="fromscratch"
             ngridk="6 6 6"
             nosource="false"
             rgkmax="8.0"
             tforce="true"
             vkloff="0 0 0"
             xctype="GGA_PBE_SOL">
        </groundstate>
    ```

    :param str xml_string: Already-prettified XML string (assumes tags are on their own lines)
    :return str reformatted_xml_string: xml_string, with the tag substrings reformatted according to the example
     - Line break per attribute.
    """
    xml_list = xml_string.split('\n')
    min_number_attributes_for_split = 3

    for i, line in enumerate(xml_list):
        if line.count("=") >= min_number_attributes_for_split:
            xml_list[i] = line_reformatter(line)

    xml_list_without_blank_lines = filter(lambda x: x.strip(), xml_list)
    return "".join(x + '\n' for x in xml_list_without_blank_lines)



""" Tests for species file class. """
import pytest

from excitingtools.species.species_file import SpeciesFile


@pytest.fixture
def species_C():
    """ Species file object for carbon. """
    species = {'chemicalSymbol': 'C', 'mass': 21.16, 'name': 'carbon', 'z': -6.0}
    muffin_tin = {'radialmeshPoints': 250, 'radius': 1.45, 'rinf': 21.09, 'rmin': 1e-05}
    atomic_states = [{'core': True, 'kappa': 1, 'l': 0, 'n': 1, 'occ': 2.0},
                     {'core': False, 'kappa': 1, 'l': 0, 'n': 2, 'occ': 2.0},
                     {'core': False, 'kappa': 1, 'l': 1, 'n': 2, 'occ': 1.0},
                     {'core': False, 'kappa': 2, 'l': 1, 'n': 2, 'occ': 1.0}]
    basis = {'default': [{'searchE': False, 'trialEnergy': 0.15, 'type': 'apw'}],
             'custom': [{'l': 0, 'n': 2, 'searchE': False, 'type': 'apw'},
                        {'l': 1, 'n': 2, 'searchE': False, 'type': 'apw'}],
             'lo': [{'l': 0, 'wf': [{'matchingOrder': 1, 'n': 2, 'searchE': False},
                                    {'matchingOrder': 2, 'n': 2, 'trialEnergy': 0.15, 'searchE': False}]},
                    {'l': 1, 'wf': [{'matchingOrder': 0, 'n': 3, 'searchE': True},
                                    {'matchingOrder': 1, 'n': 3, 'searchE': False}]}]}

    species = SpeciesFile(species, muffin_tin, atomic_states, basis)
    return species


def test_class_species_file_to_xml(species_C):
    xml_species_C = species_C.to_xml()
    assert xml_species_C.tag == "spdb", "XML root should be spdb"


def test_class_species_file_species_to_xml(species_C):
    xml_species_C = species_C.to_xml()
    sp_element = xml_species_C.find("sp")
    ref_attribs = {'chemicalSymbol': 'C', 'mass': '21.16', 'name': 'carbon', 'z': '-6.0'}
    assert sp_element.attrib == ref_attribs, "Mismatch in species attributes."


def test_class_species_file_muffin_tin_to_xml(species_C):
    xml_species_C = species_C.to_xml()
    sp_element = xml_species_C.find("sp")
    muffinTin_element = sp_element.find("muffinTin")
    ref_attribs = {'radialmeshPoints': '250', 'radius': '1.45', 'rinf': '21.09', 'rmin': '1e-05'}
    assert muffinTin_element.attrib == ref_attribs, "Mismatch in muffin-tin attributes."


def test_class_species_file_atomic_states_to_xml(species_C):
    xml_species_C = species_C.to_xml()
    sp_element = xml_species_C.find("sp")
    atomicState_elements = sp_element.findall("atomicState")
    assert atomicState_elements and len(atomicState_elements) == 4

    ref_attribs = [{'core': 'true', 'kappa': '1', 'l': '0', 'n': '1', 'occ': '2.0'},
                   {'core': 'false', 'kappa': '1', 'l': '0', 'n': '2', 'occ': '2.0'},
                   {'core': 'false', 'kappa': '1', 'l': '1', 'n': '2', 'occ': '1.0'},
                   {'core': 'false', 'kappa': '2', 'l': '1', 'n': '2', 'occ': '1.0'}]

    for i, atomicState in enumerate(atomicState_elements):
        assert atomicState.attrib == ref_attribs[
            i], f"Mismatch in atomic states at index {i}"


def test_class_species_file_basis_to_xml(species_C):
    xml_species_C = species_C.to_xml()
    sp_element = xml_species_C.find("sp")
    basis_element = sp_element.find("basis")

    default_element = basis_element.find("default")
    ref_attribs = {'searchE': 'false', 'trialEnergy': '0.15', 'type': 'apw'}
    assert default_element.attrib == ref_attribs, "Mismatch in default attributes."

    custom_elements = basis_element.findall("custom")
    assert len(
        custom_elements) == 2, f"Expected 2 custom elements, but got {len(custom_elements)}."
    ref_attribs = [{'l': '0', 'n': '2', 'searchE': 'false', 'type': 'apw'},
                   {'l': '1', 'n': '2', 'searchE': 'false', 'type': 'apw'}]
    for i, customState in enumerate(custom_elements):
        assert customState.attrib == ref_attribs[
            i], f"Mismatch in custom state at index {i}"

    lo_elements = basis_element.findall("lo")
    ref_attribs = {'lo': [{'l': '0', 'wf': [{'matchingOrder': '1', 'n': '2', 'searchE': 'false'},
                                            {'matchingOrder': '2', 'n': '2', 'trialEnergy': '0.15',
                                             'searchE': 'false'}]},
                          {'l': '1', 'wf': [{'matchingOrder': '0', 'n': '3', 'searchE': 'true'},
                                            {'matchingOrder': '1', 'n': '3', 'searchE': 'false'}]}]}
    for i, loState in enumerate(lo_elements):
        wf_elements = loState.findall('wf')
        ref_wf_attribs = ref_attribs['lo'][i]['wf']
        for j, wfState in enumerate(wf_elements):
            assert wfState.attrib == ref_wf_attribs[
                j], f"Mismatch in wf attributes at index {j} of loState at index {i}"


def test_check_matching_orders(species_C):
    species_C.basis = {'lo': [{'l': 0, 'wf': [{'matchingOrder': 1, 'n': 2, 'searchE': False},
                                              {'matchingOrder': 2, 'n': 2, 'trialEnergy': 0.15, 'searchE': False}]},
                              {'l': 3, 'wf': [{'matchingOrder': 0, 'n': 5, 'searchE': False},
                                              {'matchingOrder': 4, 'n': 5, 'searchE': False},
                                              {'matchingOrder': 3, 'n': 3, 'searchE': False}]}]}

    result_dict = species_C.check_matching_orders()
    ref_dict = {0: [2], 3: [5, 3]}
    assert result_dict == ref_dict, 'check_matching_orders failed'


def test_get_helos_from_species(species_C):
    species_C.basis = {'lo': [{'l': 0, 'wf': [{'matchingOrder': 1, 'n': 4, 'searchE': False},
                                              {'matchingOrder': 2, 'n': 2, 'trialEnergy': 0.15, 'searchE': False}]},
                              {'l': 3, 'wf': [{'matchingOrder': 2, 'n': 6, 'searchE': False},
                                              {'matchingOrder': 4, 'n': 4, 'searchE': False}]},
                              {'l': 1, 'wf': [{'matchingOrder': 0, 'n': 2, 'searchE': False},
                                              {'matchingOrder': 1, 'n': 2, 'searchE': False}]}]}

    helos_ns_per_l = species_C.get_helos_from_species()
    ref_helos_ns_per_l = {0: {4}, 1: set(), 3: {4, 6}}
    assert helos_ns_per_l == ref_helos_ns_per_l, 'get_helos_from_species failed'


def test_get_n_per_l(species_C):
    species_C.basis = {'lo': [{'l': 0, 'wf': [{'matchingOrder': 1, 'n': 4, 'searchE': False},
                                              {'matchingOrder': 2, 'n': 2, 'trialEnergy': 0.15, 'searchE': False}]},
                              {'l': 3, 'wf': [{'matchingOrder': 2, 'n': 6, 'searchE': False},
                                              {'matchingOrder': 4, 'n': 4, 'searchE': False}]},
                              {'l': 1, 'wf': [{'matchingOrder': 0, 'n': 2, 'searchE': False},
                                              {'matchingOrder': 1, 'n': 2, 'searchE': False}]}]}

    atomicstate_ns_per_l, lo_ns_per_l = species_C.get_atomicstate_ns_per_l(), species_C.get_lo_ns_per_l()

    ref_lo_ns_per_l = {0: [4, 2], 1: [2, 2], 3: [6, 4]}
    assert lo_ns_per_l == ref_lo_ns_per_l, 'get_n_per_l: get_lo_ns_per_l failed'

    ref_atomicstate_ns_per_l = {0: [1, 2], 1: [2, 2]}
    assert atomicstate_ns_per_l == ref_atomicstate_ns_per_l, 'get_n_per_l: get_atomicstate_ns_per_l failed'


def test_get_first_helo_n(species_C):
    species_C.basis = {'lo': [{'l': 0, 'wf': [{'matchingOrder': 1, 'n': 4, 'searchE': False},
                                              {'matchingOrder': 2, 'n': 2, 'trialEnergy': 0.15, 'searchE': False}]},
                              {'l': 3, 'wf': [{'matchingOrder': 2, 'n': 6, 'searchE': False},
                                              {'matchingOrder': 4, 'n': 4, 'searchE': False}]},
                              {'l': 1, 'wf': [{'matchingOrder': 0, 'n': 1, 'searchE': False},
                                              {'matchingOrder': 1, 'n': 1, 'searchE': False}]}]}

    lO_first_helo_n = species_C.get_first_helo_n(l=0)
    ref_lO_first_helo_n = 5
    assert lO_first_helo_n == ref_lO_first_helo_n, 'get_first_helo_n: for l=0 failed'

    l1_first_helo_n = species_C.get_first_helo_n(l=1)
    ref_l1_first_helo_n = 3
    assert l1_first_helo_n == ref_l1_first_helo_n, 'get_first_helo_n: for l=1 failed'

    l2_first_helo_n = species_C.get_first_helo_n(l=2)
    ref_l2_first_helo_n = 3
    assert l2_first_helo_n == ref_l2_first_helo_n, 'get_first_helo_n: for l=2 failed'

    l3_first_helo_n = species_C.get_first_helo_n(l=3)
    ref_l3_first_helo_n = 8
    assert l3_first_helo_n == ref_l3_first_helo_n, 'get_first_helo_n: for l=3 failed'


def test_add_helos(species_C):
    species_C.add_helos(0, 1)
    helo_ns_per_l = species_C.get_helos_from_species()
    ref_helo_ns_per_l = {0: {4}, 1: {3}}
    assert helo_ns_per_l == ref_helo_ns_per_l, 'add_helos: for l=0 number=1 failed'

    species_C.add_helos(3, 3)
    helo_ns_per_l = species_C.get_helos_from_species()
    ref_helo_ns_per_l = {0: {4}, 1: {3}, 3: {4, 5, 6}}
    assert helo_ns_per_l == ref_helo_ns_per_l, 'add_helos: for l=3 number=3 failed'

    species_C.add_helos(0, 2)
    helo_ns_per_l = species_C.get_helos_from_species()
    ref_helo_ns_per_l = {0: {4, 5, 6}, 1: {3}, 3: {4, 5, 6}}
    assert helo_ns_per_l == ref_helo_ns_per_l, 'add_helos: for l=0 number=2 failed'

    species_C.add_helos(1, 2)
    helo_ns_per_l = species_C.get_helos_from_species()
    ref_helo_ns_per_l = {0: {4, 5, 6}, 1: {3, 4, 5}, 3: {4, 5, 6}}
    assert helo_ns_per_l == ref_helo_ns_per_l, 'add_helos: for l=1 number=2 failed'


serialization_ref_dict = {'atomic_states': [{'core': True, 'kappa': 1, 'l': 0, 'n': 1, 'occ': 2.0},
                                            {'core': False, 'kappa': 1, 'l': 0, 'n': 2, 'occ': 2.0},
                                            {'core': False, 'kappa': 1, 'l': 1, 'n': 2, 'occ': 1.0},
                                            {'core': False, 'kappa': 2, 'l': 1, 'n': 2, 'occ': 1.0}],
                          'basis': {'custom': [{'l': 0, 'n': 2, 'searchE': False, 'type': 'apw'},
                                               {'l': 1, 'n': 2, 'searchE': False, 'type': 'apw'}],
                                    'default': [{'searchE': False, 'trialEnergy': 0.15, 'type': 'apw'}],
                                    'lo': [{'l': 0,
                                            'wf': [{'matchingOrder': 1, 'n': 2, 'searchE': False},
                                                   {'matchingOrder': 2,
                                                    'n': 2,
                                                    'searchE': False,
                                                    'trialEnergy': 0.15}]},
                                           {'l': 1,
                                            'wf': [{'matchingOrder': 0, 'n': 3, 'searchE': True},
                                                   {'matchingOrder': 1, 'n': 3, 'searchE': False}]}]},
                          'muffin_tin': {'radialmeshPoints': 250,
                                         'radius': 1.45,
                                         'rinf': 21.09,
                                         'rmin': 1e-05},
                          'species': {'chemicalSymbol': 'C', 'mass': 21.16, 'name': 'carbon', 'z': -6.0}}


def test_as_dict(species_C: SpeciesFile, mock_env_jobflow_missing):
    assert species_C.as_dict() == serialization_ref_dict, "as_dict() test failed"


def test_as_dict_jobflow(species_C: SpeciesFile, mock_env_jobflow):
    assert species_C.as_dict() == {**serialization_ref_dict, '@class': 'SpeciesFile',
                                   '@module': 'excitingtools.species.species_file'}, \
        "as_dict() with jobflow test failed"


def test_from_dict(species_C: SpeciesFile, mock_env_jobflow):
    new_species_file = species_C.from_dict(species_C.as_dict())
    assert new_species_file.species == {'chemicalSymbol': 'C', 'mass': 21.16, 'name': 'carbon', 'z': -6.0}
    assert new_species_file.muffin_tin == {'radialmeshPoints': 250, 'radius': 1.45, 'rinf': 21.09, 'rmin': 1e-05}
    assert new_species_file.atomic_states[0] == {'core': True, 'kappa': 1, 'l': 0, 'n': 1, 'occ': 2.0}
    assert set(new_species_file.basis) == {"default", "custom", "lo"}



"""Function for selecting a parser, given a file name and parsing the data.

When adding a new file, it should be added to the dictionary ` _file_to_parser`

REQUIREMENTS. Parser function must:
 a) accept a file name (not a contents string)
 b) return a dictionary.
"""
import os

from excitingtools.utils.dict_utils import container_converter

from excitingtools.exciting_dict_parsers import \
    bse_parser, groundstate_parser, gw_eigenvalues_parser, gw_eps00_parser, gw_info_parser, gw_vxc_parser, \
    input_parser, properties_parser, RT_TDDFT_parser, species_parser, state_parser


# Map file name to parser function
_file_to_parser = {
    'INFO.OUT': groundstate_parser.parse_info_out,
    'info.xml': groundstate_parser.parse_info_xml,
    'input.xml': input_parser.parse_input_xml,
    'species.xml': species_parser.parse_species_xml,
    'atoms.xml': groundstate_parser.parse_atoms,
    'evalcore.xml': groundstate_parser.parse_evalcore,
    'eigval.xml': groundstate_parser.parse_eigval,
    'geometry.xml': groundstate_parser.parse_geometry,
    'LINENGY.OUT': groundstate_parser.parse_linengy,
    'LO_RECOMMENDATION.OUT': groundstate_parser.parse_lo_recommendation,
    'RHO3D.xml': properties_parser.parse_plot_3d,
    'VCL3D.xml': properties_parser.parse_plot_3d,
    'VXC3D.xml': properties_parser.parse_plot_3d,
    'WF3D.xml': properties_parser.parse_plot_3d,
    'ELF3D.xml': properties_parser.parse_plot_3d,
    'EF3D.xml': properties_parser.parse_plot_3d,
    'LSJ.xml': properties_parser.parse_lsj,
    'EFG.xml': properties_parser.parse_efg,
    'mossbauer.xml': properties_parser.parse_mossbauer,
    'expiqr.xml': properties_parser.parse_expiqr,
    'effmass.xml': properties_parser.parse_effmass,
    'bandstructure.xml': properties_parser.parse_bandstructure_depreciated,
    'dos.xml': properties_parser.parse_dos,
    'KERR.OUT': properties_parser.parse_kerr,
    'EPSILON_11.OUT': properties_parser.parse_epsilon,
    'EPSILON_12.OUT': properties_parser.parse_epsilon,
    'EPSILON_33.OUT': properties_parser.parse_epsilon,
    'CHI_111.OUT': properties_parser.parse_chi,
    'ELNES.OUT': properties_parser.parse_elnes,
    'SEEBECK_11.OUT': properties_parser.parse_seebeck,
    'ELECTCOND_11.OUT': properties_parser.parse_seebeck,
    'THERMALCOND_11.OUT': properties_parser.parse_seebeck,
    'Z_11.OUT': properties_parser.parse_seebeck,
    'ldos.out': properties_parser.parse_ldos,
    'band_edges.out': properties_parser.parse_band_edges,
    'spintext.xml': properties_parser.parse_spintext,
    'POLARIZATION.OUT': properties_parser.parse_polarization,
    'TDOS_WANNIER.OUT': properties_parser.parse_tdos_wannier,
    'WANNIER_INFO.OUT': properties_parser.parse_wannier_info,
    'coreoverlap.xml': properties_parser.parse_core_overlap,
    'INFOXS.OUT': bse_parser.parse_infoxs_out,
    'EPSILON_NAR_BSE-singlet-TDA-BAR_SCR-full_OC11.OUT': bse_parser.parse_EPSILON_NAR,
    'EPSILON_NAR_BSE-singlet-TDA-BAR_SCR-full_OC22.OUT': bse_parser.parse_EPSILON_NAR,
    'EPSILON_NAR_BSE-singlet-TDA-BAR_SCR-full_OC33.OUT': bse_parser.parse_EPSILON_NAR,
    'EPSILON_NAR_FXCMB1_OC11_QMT001.OUT': bse_parser.parse_EPSILON_NAR,
    'EPSILON_NAR_FXCMB1_OC22_QMT001.OUT': bse_parser.parse_EPSILON_NAR,
    'EPSILON_NAR_FXCMB1_OC33_QMT001.OUT': bse_parser.parse_EPSILON_NAR,
    'EPSILON_NAR_NLF_FXCMB1_OC11_QMT001.OUT': bse_parser.parse_EPSILON_NAR,
    'EPSILON_NAR_NLF_FXCMB1_OC22_QMT001.OUT': bse_parser.parse_EPSILON_NAR,
    'EPSILON_NAR_NLF_FXCMB1_OC33_QMT001.OUT': bse_parser.parse_EPSILON_NAR,
    'EPSILON_BSE-singlet-TDA-BAR_SCR-full_OC11.OUT': bse_parser.parse_EPSILON_NAR,
    'EPSILON_BSE-singlet-TDA-BAR_SCR-full_OC22.OUT': bse_parser.parse_EPSILON_NAR,
    'EPSILON_BSE-singlet-TDA-BAR_SCR-full_OC33.OUT': bse_parser.parse_EPSILON_NAR,
    'EPSILON_BSE-singlet-TDA-BAR_SCR-full_OC12.OUT': bse_parser.parse_EPSILON_NAR,
    'DICHROIC_K_BSE-singlet-TDA-BAR_SCR-full_OC11.OUT': bse_parser.parse_EPSILON_NAR,
    'DICHROIC_K_BSE-singlet-TDA-BAR_SCR-full_OC22.OUT': bse_parser.parse_EPSILON_NAR,
    'DICHROIC_K_BSE-singlet-TDA-BAR_SCR-full_OC33.OUT': bse_parser.parse_EPSILON_NAR,
    'DICHROIC_K_BSE-singlet-TDA-BAR_SCR-full_OC12.OUT': bse_parser.parse_EPSILON_NAR,
    'DICHROIC_KBAR_BSE-singlet-TDA-BAR_SCR-full_OC11.OUT': bse_parser.parse_EPSILON_NAR,
    'DICHROIC_KBAR_BSE-singlet-TDA-BAR_SCR-full_OC22.OUT': bse_parser.parse_EPSILON_NAR,
    'DICHROIC_KBAR_BSE-singlet-TDA-BAR_SCR-full_OC33.OUT': bse_parser.parse_EPSILON_NAR,
    'DICHROIC_KBAR_BSE-singlet-TDA-BAR_SCR-full_OC12.OUT': bse_parser.parse_EPSILON_NAR,
    'DICHROIC_K_KBAR_BSE-singlet-TDA-BAR_SCR-full_OC11.OUT': bse_parser.parse_EPSILON_NAR,
    'DICHROIC_K_KBAR_BSE-singlet-TDA-BAR_SCR-full_OC22.OUT': bse_parser.parse_EPSILON_NAR,
    'DICHROIC_K_KBAR_BSE-singlet-TDA-BAR_SCR-full_OC33.OUT': bse_parser.parse_EPSILON_NAR,
    'DICHROIC_K_KBAR_BSE-singlet-TDA-BAR_SCR-full_OC12.OUT': bse_parser.parse_EPSILON_NAR,
    'OSCI_K_BSE-singlet-TDA-BAR_SCR-full_OC11.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'OSCI_K_BSE-singlet-TDA-BAR_SCR-full_OC22.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'OSCI_K_BSE-singlet-TDA-BAR_SCR-full_OC33.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'OSCI_K_BSE-singlet-TDA-BAR_SCR-full_OC12.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'OSCI_KBAR_BSE-singlet-TDA-BAR_SCR-full_OC11.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'OSCI_KBAR_BSE-singlet-TDA-BAR_SCR-full_OC22.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'OSCI_KBAR_BSE-singlet-TDA-BAR_SCR-full_OC33.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'OSCI_KBAR_BSE-singlet-TDA-BAR_SCR-full_OC12.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'EPSILON_BSE-IP_SCR-full_OC11.OUT': bse_parser.parse_EPSILON_NAR,
    'EPSILON_BSE-IP_SCR-full_OC22.OUT': bse_parser.parse_EPSILON_NAR,
    'EPSILON_BSE-IP_SCR-full_OC33.OUT': bse_parser.parse_EPSILON_NAR,
    'LOSS_NAR_FXCMB1_OC11_QMT001.OUT': bse_parser.parse_LOSS_NAR,
    'LOSS_NAR_FXCMB1_OC22_QMT001.OUT': bse_parser.parse_LOSS_NAR,
    'LOSS_NAR_FXCMB1_OC33_QMT001.OUT': bse_parser.parse_LOSS_NAR,
    'LOSS_NAR_NLF_FXCMB1_OC11_QMT001.OUT': bse_parser.parse_LOSS_NAR,
    'LOSS_NAR_NLF_FXCMB1_OC22_QMT001.OUT': bse_parser.parse_LOSS_NAR,
    'LOSS_NAR_NLF_FXCMB1_OC33_QMT001.OUT': bse_parser.parse_LOSS_NAR,
    'LOSS_BSE-singlet-TDA-BAR_SCR-full_OC11.OUT': bse_parser.parse_LOSS_NAR,
    'LOSS_BSE-singlet-TDA-BAR_SCR-full_OC22.OUT': bse_parser.parse_LOSS_NAR,
    'LOSS_BSE-singlet-TDA-BAR_SCR-full_OC33.OUT': bse_parser.parse_LOSS_NAR,
    'LOSS_BSE-IP_SCR-full_OC11.OUT': bse_parser.parse_LOSS_NAR,
    'LOSS_BSE-IP_SCR-full_OC22.OUT': bse_parser.parse_LOSS_NAR,
    'LOSS_BSE-IP_SCR-full_OC33.OUT': bse_parser.parse_LOSS_NAR,
    'EXCITON_NAR_BSE-singlet-TDA-BAR_SCR-full_OC11.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'EXCITON_NAR_BSE-singlet-TDA-BAR_SCR-full_OC22.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'EXCITON_NAR_BSE-singlet-TDA-BAR_SCR-full_OC33.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'EXCITON_BSE-singlet-TDA-BAR_SCR-full_OC11.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'EXCITON_BSE-singlet-TDA-BAR_SCR-full_OC22.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'EXCITON_BSE-singlet-TDA-BAR_SCR-full_OC33.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'EXCITON_BSE-IP_SCR-full_OC11.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'EXCITON_BSE-IP_SCR-full_OC22.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'EXCITON_BSE-IP_SCR-full_OC33.OUT': bse_parser.parse_EXCITON_NAR_BSE,
    'GW_INFO.OUT': gw_info_parser.parse_gw_info,
    'EFERMI_GW.OUT': gw_eigenvalues_parser.parse_efermi_gw,
    'EVALQP.DAT': gw_eigenvalues_parser.parse_evalqp,
    'VXCNN.DAT': gw_vxc_parser.parse_vxcnn,
    'EPS00_GW.OUT': gw_eps00_parser.parse_eps00_gw,
    'JIND.OUT': RT_TDDFT_parser.parse_jind,
    'NEXC.OUT': RT_TDDFT_parser.parse_nexc,
    'ETOT_RTTDDFT.OUT': RT_TDDFT_parser.parse_etot,
    'EIGVAL_': RT_TDDFT_parser.parse_eigval_screenshots,
    'PROJ_': RT_TDDFT_parser.parse_proj_screenshots,
    'ATOM_': RT_TDDFT_parser.parse_atom_position_velocity_force,
    'FCR_': RT_TDDFT_parser.parse_force,
    'FEXT_': RT_TDDFT_parser.parse_force,
    'FHF_': RT_TDDFT_parser.parse_force,
    'FVAL_': RT_TDDFT_parser.parse_force,
    'wf1d-0001-0001.dat': properties_parser.parse_wf1d,
    'wf1d-0003-0001.dat': properties_parser.parse_wf1d,
    'wf2d-0001-0001.xsf': properties_parser.parse_wf2d,
    'wf2d-0003-0001.xsf': properties_parser.parse_wf2d,
    'wf3d-0001-0001.xsf': properties_parser.parse_wf3d,
    'wf3d-0003-0001.xsf': properties_parser.parse_wf3d,
    'wf3d-0001-0001.cube': properties_parser.parse_cube,
    'wf3d-0003-0001.cube': properties_parser.parse_cube,
    'STATE.OUT': state_parser.parse_state_out,
}


def truncate_fnames_with_exts(file_name: str) -> str:
    """ Truncate file names that have open-ended extensions.

    For example:
      EIGVAL_00.dat -> EIGVAL_
      EIGVAL_01.dat -> EIGVAL_

    :param file_name: File name containing fixed prefix and
    an extension beginning with '_'.
    :return file_name: File name prefix, else input file name.
    """
    prefixes = [
        'EIGVAL_',
        'PROJ_',
        'ATOM_',
        'FCR_',
        'FEXT_',
        'FHF_',
        'FVAL_'
    ]
    if any( [ prefix in file_name for prefix in prefixes ] ) :
        file_name_prefix = file_name.split('_')[0] + '_'
        return file_name_prefix

    return file_name


def parser_chooser(full_file_name: str) -> dict:
    """ Selects parser according to the name of the input file then returns the result of the parser.

    REQUIREMENTS. Parser function must:
     a) accept a file name (not a contents string)
     b) return a dictionary.

    param: str, full_file_name: file name prepended by full path
    return: parsed data
    """
 
    full_file_name = full_file_name.rstrip()
    if not os.path.exists(full_file_name):
        raise FileNotFoundError(f'File not found: {full_file_name}')

    file_name = os.path.split(full_file_name)[1]
    file_name = truncate_fnames_with_exts(file_name)

    files_with_parsers = [name for name in _file_to_parser.keys()]
    if file_name not in files_with_parsers:
        raise KeyError(f"File does not have a parser: {file_name}")

    parser = _file_to_parser[file_name]
    data = parser(full_file_name)

    #  TODO(Alex) Issue 135 Ensure all parsers return appropriate values, not strings.
    #   container_converter should therefore be used as a decorator on parsers with values that are strings.
    #   That will massively speed up parsing
    return container_converter(data)




""" Data Structures.

Data structure is defined as a container for data.
Many of these classes could be @dataclass, however excitingtools
retains support for python 3.6.
"""

class PointIndex:
    """ Container for (point, index) pair
    """
    def __init__(self, point, index: int):
        self.point = point
        self.index = index

class BandIndices:
    """Indices of valence band maximum and conduction band minimum"""
    def __init__(self, VBM: int, CBm: int):
        self.VBM = VBM
        self.CBm = CBm

class NumberOfStates:
    """Number of states. Useful when indexing does not start at 0/1
    """
    def __init__(self, first_state: int, last_state: int):
        self.first_state = first_state
        self.last_state = last_state
        self.n_states = last_state - first_state + 1



"""Tests for the binary runner.
Excludes the run method.
"""
import shutil
import sys
from pathlib import Path
from typing import Any

import pytest

from excitingtools.runner.runner import BinaryRunner, RunnerCode

mock_binary = "false_exciting_binary"


@pytest.mark.xfail(shutil.which(mock_binary) is not None, reason="Binary name exists.")
def test_no_binary():
    with pytest.raises(FileNotFoundError,
                       match=fr"{mock_binary} binary is not present in the current directory nor in \$PATH"):
        BinaryRunner(mock_binary, "./", 1, 1)


@pytest.fixture
def exciting_smp(tmp_path: Path) -> str:
    binary = tmp_path / "exciting_smp"
    binary.touch()
    return binary.as_posix()


def test_no_run_dir(exciting_smp: str):
    with pytest.raises(OSError, match="Run directory does not exist: non_existent_dir"):
        BinaryRunner(exciting_smp, "./", 1, 1, "non_existent_dir")


def test_false_run_cmd(exciting_smp: str):
    false_run_cmd: Any = 3
    with pytest.raises(ValueError, match="Run commands expected in a str or list. For example ['mpirun', '-np', '2']"):
        BinaryRunner(exciting_smp, false_run_cmd, 1, 1)


@pytest.fixture
def exciting_mpismp(tmp_path: Path) -> str:
    binary = tmp_path / "exciting_mpismp"
    binary.touch()
    return binary.as_posix()


def test_false_mpi_command_smaller_than_zero(exciting_mpismp: str):
    with pytest.raises(ValueError, match="Number of MPI processes must be > 0"):
        BinaryRunner(exciting_mpismp, ["mpirun", "-np", "-1"], 1, 1)


def test_false_mpi_command_no_int(exciting_mpismp: str):
    with pytest.raises(ValueError, match="Number of MPI processes should be an int"):
        BinaryRunner(exciting_mpismp, ["mpirun", "-np", "no_int"], 1, 1)


def test_false_mpi_command_no_number_given(exciting_mpismp: str):
    with pytest.raises(ValueError, match="Number of MPI processes must be specified after the '-np'"):
        BinaryRunner(exciting_mpismp, ["mpirun", "-np"], 1, 1)


def test_false_omp(exciting_smp: str):
    with pytest.raises(ValueError, match="Number of OMP threads must be > 0"):
        BinaryRunner(exciting_smp, [""], -1, 1)


def test_false_timeout(exciting_smp: str):
    with pytest.raises(ValueError, match="time_out must be a positive integer"):
        BinaryRunner(exciting_smp, [""], 1, -1)


@pytest.fixture
def runner(tmp_path: Path, exciting_mpismp: str) -> BinaryRunner:
    """Produces a runner with binary and run dir mocked up.
    """
    run_dir = tmp_path / "ab/de"
    run_dir.mkdir(parents=True)
    return BinaryRunner(exciting_mpismp, ["mpirun", "-np", "3"], 4, 260, run_dir.as_posix(), [">", "std.out"])


def test_as_dict(tmp_path: Path, runner: BinaryRunner, mock_env_jobflow_missing):
    assert runner.as_dict() == {'args': ['>', 'std.out'],
                                'binary': (tmp_path / "exciting_mpismp").as_posix(),
                                'directory': (tmp_path / "ab/de").as_posix(),
                                'omp_num_threads': 4,
                                'run_cmd': ['mpirun', '-np', '3'],
                                'time_out': 260}


def test_as_dict_jobflow(tmp_path: Path, runner: BinaryRunner, mock_env_jobflow):
    assert runner.as_dict() == {'@class': 'BinaryRunner',
                                '@module': 'excitingtools.runner.runner',
                                'args': ['>', 'std.out'],
                                'binary': (tmp_path / "exciting_mpismp").as_posix(),
                                'directory': (tmp_path / "ab/de").as_posix(),
                                'omp_num_threads': 4,
                                'run_cmd': ['mpirun', '-np', '3'],
                                'time_out': 260}


def test_from_dict(tmp_path: Path, runner):
    new_runner = BinaryRunner.from_dict(runner.as_dict())
    assert new_runner.binary == (tmp_path / "exciting_mpismp").as_posix()
    assert new_runner.time_out == 260


def test_run_with_bash_command(tmp_path: Path):
    """Produces a runner with binary and run dir mocked up.
    Test a simple echo command.
    """
    run_dir = tmp_path / "ab/de"
    run_dir.mkdir(parents=True)
    binary = tmp_path / "hello"
    binary.touch()
    runner = BinaryRunner(binary.as_posix(), ["echo"], 1, 60, run_dir.as_posix())
    run_results = runner.run()
    assert run_results.success
    assert run_results.stderr == b''
    assert run_results.stdout.decode() == binary.as_posix() + "\n"


@pytest.mark.skipif(sys.version_info < (3, 8), reason="requires python3.8 or higher")
def test_timeout_with_bash_command(tmp_path: Path):
    """Produces a runner with binary and run dir mocked up.
    Test a simple sleep command to get a timeout.

    Note: Skip for python3.7 because it returns b'' as stdout instead of None
    """
    time_out = 1
    binary = tmp_path / "sleep.sh"
    binary.write_text(f"sleep {time_out + 0.1}")
    runner = BinaryRunner(binary.as_posix(), ["sh"], 1, time_out, tmp_path.as_posix())
    run_results = runner.run()
    assert not run_results.success
    assert run_results.stderr == 'BinaryRunner: Job timed out. \n\n'
    assert run_results.stdout is None
    assert run_results.process_time == time_out
    assert isinstance(run_results.return_code, RunnerCode)
    assert run_results.return_code == RunnerCode.time_out



""" Test all GW output file parsers, except GW_INFO.OUT
"""
import pytest
import numpy as np

from excitingtools.utils.test_utils import MockFile
from excitingtools.utils.utils import get_new_line_indices

from excitingtools.exciting_dict_parsers.gw_eps00_parser import parse_eps00_frequencies, parse_eps00_gw, \
    _file_name


@pytest.fixture
def eps00_mock(tmp_path):
    # EPS00_GW.OUT segment. Note that first line of file must be blank.
    eps_string = """
(dielectric tensor, random phase approximation)

frequency index and value:      1    0.00529953
real part, imaginary part below
   8.31881773    0.00000000    0.00000000         0.00000000    0.00000000    0.00000000
   0.00000000    8.31881773    0.00000000         0.00000000    0.00000000    0.00000000
   0.00000000    0.00000000    8.31881773         0.00000000    0.00000000    0.00000000

frequency index and value:      2    0.02771249
real part, imaginary part below
   8.22189228    0.00000000    0.00000000        -0.00000000    0.00000000    0.00000000
   0.00000000    8.22189228    0.00000000         0.00000000   -0.00000000    0.00000000
   0.00000000    0.00000000    8.22189228         0.00000000    0.00000000   -0.00000000

frequency index and value:      3    0.06718440
real part, imaginary part below
   7.78004308    0.00000000    0.00000000         0.00000000    0.00000000    0.00000000
   0.00000000    7.78004308    0.00000000         0.00000000    0.00000000    0.00000000
   0.00000000    0.00000000    7.78004308         0.00000000    0.00000000    0.00000000

"""
    eps00_file = tmp_path / _file_name
    eps00_file.write_text(eps_string)
    return MockFile(eps00_file, eps_string)


def test_parse_eps00_frequencies(eps00_mock):
    """ Test parsing frequencies from EPS00_GW.OUT
    """
    line = get_new_line_indices(eps00_mock.string)

    ref = {1: 0.00529953, 2: 0.02771249, 3: 0.06718440}

    assert eps00_mock.string[line[0]:line[1]].isspace(), "First line of eps_string must be a whiteline"
    assert parse_eps00_frequencies(eps00_mock.string) == ref, "Frequency grid for eps00"


def test_parse_eps00_gw(eps00_mock):
    """ Test parsing EPS00_GW.OUT
    """
    line = get_new_line_indices(eps00_mock.string)
    assert eps00_mock.string[line[0]:line[1]].isspace(), "First line of eps_string must be a whiteline"

    ref = {
        1: {
            'frequency': 0.00529953,
            'eps00': {
                're': np.array([[8.31881773, 0., 0.], [0., 8.31881773, 0.], [0., 0., 8.31881773]]),
                'img': np.array([[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]])
                }
            },
        2: {
            'frequency': 0.02771249,
            'eps00': {
                're': np.array([[8.22189228, 0., 0.], [0., 8.22189228, 0.], [0., 0., 8.22189228]]),
                'img': np.array([[-0., 0., 0.], [0., -0., 0.], [0., 0., -0.]])
                }
            },
        3: {
            'frequency': 0.06718440,
            'eps00': {
                're': np.array([[7.78004308, 0.0, 0.0], [0., 7.78004308, 0.], [0., 0., 7.78004308]]),
                'img': np.array([[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]])
                }
            }
        }

    output = parse_eps00_gw(eps00_mock.file)

    assert len(output) == 3, "3 frequency points"
    assert [k for k in output[1].keys()] == ['frequency', 'eps00'], "Frequency point 1 keys "
    assert [k for k in output[2].keys()] == ['frequency', 'eps00'], "Frequency point 2 keys "
    assert [k for k in output[3].keys()] == ['frequency', 'eps00'], "Frequency point 3 keys "

    assert output[1]['frequency'] == 0.00529953, "Frequency point 1 value"
    assert output[2]['frequency'] == 0.02771249, "Frequency point 2 value"
    assert output[3]['frequency'] == 0.06718440, "Frequency point 3 value"

    assert np.allclose(output[1]['eps00']['re'], ref[1]['eps00']['re']),"Re{eps00} at frequency point 1"
    assert np.allclose(output[1]['eps00']['img'], ref[1]['eps00']['img']),"Im{eps00} at frequency point 1"

    assert np.allclose(output[2]['eps00']['re'], ref[2]['eps00']['re']),"Re{eps00} at frequency point 2"
    assert np.allclose(output[2]['eps00']['img'], ref[2]['eps00']['img']),"Im{eps00} at frequency point 2"

    assert np.allclose(output[3]['eps00']['re'], ref[3]['eps00']['re']),"Re{eps00} at frequency point 3"
    assert np.allclose(output[3]['eps00']['img'], ref[3]['eps00']['img']),"Im{eps00} at frequency point 3"



"""Test ExcitingGroundStateInput class attribute assignment and methods.

NOTE:
All attribute tests should assert on the XML tree content's as the attribute
order is not preserved by the ElementTree.tostring method. Elements appear to
be fine.

For example:

 `gs_xml_string = xml.etree.ElementTree.tostring(
     gs_xml, encoding='unicode', method='xml')`

may return:

'<groundstate ngridk="8 8 8" rgkmax="8.6"> </groundstate>'
or
'<groundstate rgkmax="8.6" ngridk="8 8 8"> </groundstate>'

"""
import pytest

from excitingtools.input.input_classes import ExcitingGroundStateInput


@pytest.mark.parametrize(
    "test_input,expected",
    [({"rgkmax": 8.6}, [('rgkmax', '8.6')]), ({"ngridk": [8, 8, 8]}, [('ngridk', '8 8 8')]),
     ({'vkloff': [0.1, 0.2, 0.3]}, [('vkloff', '0.1 0.2 0.3')]),
     ({'CoreRelativity': 'dirac'}, [('CoreRelativity', 'dirac')]),
     ({'ExplicitKineticEnergy': True}, [('ExplicitKineticEnergy', 'true')]),
     ({'PrelimLinSteps': 2}, [('PrelimLinSteps', '2')]),
     ({'ValenceRelativity': 'zora'}, [('ValenceRelativity', 'zora')]),
     ({'autokpt': False}, [('autokpt', 'false')]), ({'beta0': 0.4}, [('beta0', '0.4')]),
     ({'betadec': 0.6}, [('betadec', '0.6')]), ({'betainc': 1.1}, [('betainc', '1.1')]),
     ({'cfdamp': 0.0}, [('cfdamp', '0.0')]), ({'chgexs': 0.0}, [('chgexs', '0.0')]),
     ({'deband': 2.5e-3}, [('deband', '0.0025')]),
     ({'dipolecorrection': True}, [('dipolecorrection', 'true')]),
     ({'dipoleposition': 1.0}, [('dipoleposition', '1.0')]),
     ({'dlinengyfermi': -0.1}, [('dlinengyfermi', '-0.1')]),
     ({'do': "fromscratch"}, [('do', "fromscratch")]),
     ({'energyref': 0.0}, [('energyref', '0.0')]), ({'epsband': 1.0e-6}, [('epsband', '1e-06')]),
     ({'epschg': 1.0e-5}, [('epschg', '1e-05')]), ({'epsengy': 1e-6}, [('epsengy', '1e-06')]),
     ({'epsforcescf': 5.0e-5}, [('epsforcescf', '5e-05')]),
     ({'epsocc': 1e-8}, [('epsocc', '1e-08')]), ({'epspot': 1e-6}, [('epspot', '1e-06')]),
     ({'fermilinengy': False}, [('fermilinengy', 'false')]),
     ({'findlinentype': "Wigner_Seitz"}, [('findlinentype', "Wigner_Seitz")]),
     ({'fracinr': 0.02}, [('fracinr', '0.02')]), ({'frozencore': False}, [('frozencore', 'false')]),
     ({'gmaxvr': 12}, [('gmaxvr', '12')]), ({'isgkmax': -1}, [('isgkmax', '-1')]),
     ({'ldapu': "none"}, [('ldapu', "none")]), ({'lmaxapw': 8}, [('lmaxapw', '8')]),
     ({'lmaxinr': 2}, [('lmaxinr', '2')]), ({'lmaxmat': 8}, [('lmaxmat', '8')]),
     ({'lmaxvr': 8}, [('lmaxvr', '8')]), ({'lradstep': 1}, [('lradstep', '1')]),
     ({'maxscl': 200}, [('maxscl', '200')]), ({'mixer': 'msec'}, [('mixer', 'msec')]),
     ({'mixerswitch': 1}, [('mixerswitch', '1')]), ({'modifiedsv': False}, [('modifiedsv', 'false')]),
     ({'msecStoredSteps': 8}, [('msecStoredSteps', '8')]), ({'nempty': 5}, [('nempty', '5')]),
     ({'niterconvcheck': 2}, [('niterconvcheck', '2')]), ({'nktot': 0}, [('nktot', '0')]),
     ({'nosource': False}, [('nosource', 'false')]), ({'nosym': False}, [('nosym', 'false')]),
     ({'nprad': 4}, [('nprad', '4')]), ({'npsden': 9}, [('npsden', '9')]),
     ({'nwrite': 0}, [('nwrite', '0')]), ({'outputlevel': 'normal'}, [('outputlevel', 'normal')]),
     ({'ptnucl': True}, [('ptnucl', 'true')]), ({'radialgridtype': 'cubic'}, [('radialgridtype', 'cubic')]),
     ({'radkpt': 40.0}, [('radkpt', '40.0')]), ({'reducek': True}, [('reducek', 'true')]),
     ({'scfconv': 'multiple'}, [('scfconv', 'multiple')]), ({'stype': 'Gaussian'}, [('stype', 'Gaussian')]),
     ({'swidth': 0.001}, [('swidth', '0.001')]), ({'symmorph': False}, [('symmorph', 'false')]),
     ({'tevecsv': False}, [('tevecsv', 'false')]), ({'tfibs': True}, [('tfibs', 'true')]),
     ({'tforce': False}, [('tforce', 'false')]), ({'tpartcharges': False}, [('tpartcharges', 'false')]),
     ({'useDensityMatrix': True}, [('useDensityMatrix', 'true')]),
     ({'vdWcorrection': 'none'}, [('vdWcorrection', 'none')]),
     ({'xctype': 'GGA_PBE'}, [('xctype', 'GGA_PBE')])]
)
def test_class_exciting_ground_state_input_parametrized(test_input, expected):
    gs_input = ExcitingGroundStateInput(**test_input)
    gs_xml = gs_input.to_xml()
    assert gs_xml.tag == 'groundstate'
    assert gs_xml.items() == expected


def test_invalid_input():
    """
    Test error is raised when giving bogus attributes to class constructor.
    """
    # Use an erroneous ground state attribute
    with pytest.raises(ValueError, match="groundstate keys are not valid: {'erroneous_attribute'}"):
        ExcitingGroundStateInput(erroneous_attribute=True)


def test_as_dict(mock_env_jobflow_missing):
    ref_rgkmax = 8.5
    gs_input = ExcitingGroundStateInput(rgkmax=ref_rgkmax)
    ref_dict = {'xml_string': f'<groundstate rgkmax="{ref_rgkmax}"> </groundstate>'}
    assert gs_input.as_dict() == ref_dict, 'expected different dict representation'


def test_as_dict_jobflow(mock_env_jobflow):
    ref_rgkmax = 8.5
    gs_input = ExcitingGroundStateInput(rgkmax=ref_rgkmax)
    ref_dict = {'@class': 'ExcitingGroundStateInput',
                '@module': 'excitingtools.input.input_classes',
                'xml_string': f'<groundstate rgkmax="{ref_rgkmax}"> </groundstate>'}
    assert gs_input.as_dict() == ref_dict, 'expected different dict representation'


def test_from_dict():
    ref_rgkmax = 8.5
    ref_dict = {'xml_string': f'<groundstate rgkmax="{ref_rgkmax}"> </groundstate>'}
    gs_input = ExcitingGroundStateInput.from_dict(ref_dict)

    assert gs_input.name == 'groundstate'
    # added comment for pylint to disable warning, because of dynamic attributes
    assert gs_input.rgkmax == ref_rgkmax, f'Expect rgkmax to be equal {ref_rgkmax}'  # pylint: disable=no-member


def test_spin_input():
    spin_attributes = {'bfieldc': [0, 0, 0], 'fixspin': 'total FSM'}
    spin_keys = list(spin_attributes)
    gs_input = ExcitingGroundStateInput(rgkmax=7.0, spin=spin_attributes)

    gs_xml = gs_input.to_xml()
    assert gs_xml.tag == 'groundstate'
    assert set(gs_xml.keys()) == {'rgkmax'}

    elements = list(gs_xml)
    assert len(elements) == 1

    spin_xml = elements[0]
    assert spin_xml.tag == "spin"
    assert spin_xml.keys() == spin_keys, 'Should contain all spin attributes'
    assert spin_xml.get('bfieldc') == '0 0 0'
    assert spin_xml.get('fixspin') == 'total FSM'


def test_solver_input():
    solver_attributes = {'packedmatrixstorage': True, 'type': "Lapack"}
    solver_keys = list(solver_attributes)
    gs_input = ExcitingGroundStateInput(solver=solver_attributes)

    gs_xml = gs_input.to_xml()
    assert gs_xml.tag == 'groundstate'
    assert set(gs_xml.keys()) == set()

    elements = list(gs_xml)
    assert len(elements) == 1

    solver_xml = elements[0]
    assert solver_xml.tag == "solver"
    assert solver_xml.keys() == solver_keys, 'Should contain all spin attributes'
    assert solver_xml.get('packedmatrixstorage') == 'true'
    assert solver_xml.get('type') == 'Lapack'


def test_dfthalf_input():
    dfthalf_attributes = {"printVSfile": True}
    dfthalf_keys = list(dfthalf_attributes)
    gs_input = ExcitingGroundStateInput(dfthalf=dfthalf_attributes)

    gs_xml = gs_input.to_xml()
    assert gs_xml.tag == 'groundstate'
    assert set(gs_xml.keys()) == set()

    elements = list(gs_xml)
    assert len(elements) == 1

    dfthalf_xml = elements[0]
    assert dfthalf_xml.tag == "dfthalf"
    assert dfthalf_xml.keys() == dfthalf_keys, 'Should contain all dfthalf attributes'
    assert dfthalf_xml.get('printVSfile') == 'true'


def test_OEP_input():
    oep_attributes = {"convoep": 1e-5, "maxitoep": 200, "tauoep": [1, 2, 3]}
    oep_keys = list(oep_attributes)
    gs_input = ExcitingGroundStateInput(OEP=oep_attributes)

    gs_xml = gs_input.to_xml()
    assert gs_xml.tag == 'groundstate'
    assert set(gs_xml.keys()) == set()

    elements = list(gs_xml)
    assert len(elements) == 1

    oep_xml = elements[0]
    assert oep_xml.tag == "OEP"
    assert oep_xml.keys() == oep_keys, 'Should contain all dfthalf attributes'
    assert oep_xml.get('convoep') == '1e-05'
    assert oep_xml.get('maxitoep') == '200'
    assert oep_xml.get('tauoep') == '1 2 3'



""" Parse GW's EVALQP.DAT file into dicts.
Also seemed like a logical place to add the Fermi-level parser.
"""
import enum
from itertools import count
import numpy as np
import re
import os

from excitingtools.dataclasses.data_structs import NumberOfStates


# GW eigenvalue file name
_file_name = 'EVALQP.DAT'


def parse_efermi_gw(name: str) -> dict:
    """ Parser for EFERMI_GW.OUT.

    :param name: File name
    :return data: GW Fermi level.
    """
    data = {}
    try:
        data["EFERMI_GW"] = np.genfromtxt(name)
    except Exception:
        raise Exception("Numpy parsing error")
    return data


def k_points_from_evalqp(file_string: str) -> dict:
    """Get the irreducible k/q-points and weights from 'EVALQP.DAT'

    These *can*  differ from those reported in 'KPOINTS', depending
    on the choice of ngridq in GW

    :param file_string: File string.
    :return k_points: k-points and their weights.
    """
    k_points_raw: list = re.findall(r'\s*k-point .*$', file_string, flags=re.MULTILINE)
    k_points = {}
    for ik, line in enumerate(k_points_raw):
        kx, ky, kz, w = line.split()[-4:]
        k_points[ik + 1] = {
            'k_point': [float(k) for k in [kx, ky, kz]],
            'weight': float(w)
        }
    return k_points


def n_states_from_evalqp(file_string: str) -> NumberOfStates:
    return n_states_from_file(file_string, n_header=2)


def n_states_from_file(file_string: str, n_header: int) -> NumberOfStates:
    """Get the total number of states used per k-point, from EVALQP.DAT.

    Expect n_states to be fixed per k-point, however GW corrections can apply
    over any state interval [ibgw: nbgw], therefore extract the first state
    index and the last state index from the first and last k-points,
    respectively - for optimal parsing efficiency.

    Could also extract from input.xml.

    :param str file_string: Input string.
    :param n_header Number of header lines.
    :return int n_states: Number of states per k-point (occupied plus empty).
    """
    lines = file_string.splitlines()

    # ibgw = first_state
    first_state = int(lines[n_header].split()[0])

    # nbgw = last_state
    last_state = None
    for line in reversed(lines):
        if len(line.strip()) > 0:
            last_state = int(line.split()[0])
            break

    assert last_state is not None, "Index of final state to have GW correct applied not found"

    return NumberOfStates(first_state, last_state)


def parse_evalqp_blocks(full_file_name: str, k_points: dict, n_states: int) -> dict:
    """Parse energy information from EVALQP.dat.

    The function expects k-points of the form:

      k_points[ik] = {'k_point': k_point, 'weight': weight}

    where the k-index (ik) follows fortran indexing convention, and is expected to be
    contiguous. The function returns parsed data, with each element of shape (n_states, 10).

    The routine exploits the repeating structure EVALQP.dat:
       kpoint k1 k2 k3 weight
       header line
       first state (ibgw)
       .
       .
       last state (nbgw)

       kpoint k1 k2 k3 weight
       header line
       first state (ibgw)
       .
       .
       last state (nbgw)

    to parse all energies per k-point.

    :param str full_file_name: Path + file name
    :param dict k_points: Dictionary of k-points
    :param int n_states: Total number of occupied plus empty states
     Note, this is constant per q-point.

    :return dict data: Parsed energies from EVALQP.dat
    """
    # File formatting
    header_size = 2
    blank_line = 1

    data = {}
    skip_lines = header_size

    # Must iterate lowest to highest, else data won't match k-points
    for ik in range(1, len(k_points) + 1):
        block_data = np.loadtxt(full_file_name,
                                skiprows=skip_lines,
                                max_rows=n_states)
        # Ignore first column (state index)
        data[ik] = block_data[:, 1:]
        skip_lines += n_states + (header_size + blank_line)

    return data


def parse_evalqp(full_file_name: str) -> dict:
    """Parse GW output file EVALQP.DAT

    Parse  and return data of the form:
      data[ik] = {'k_point': k_point, 'weight': weight, 'energies': energies}

    NOTE(Alex) Would be good to transpose this: row-major access.

    For oxygen release:
      energies have the shape (n_states, 10), where the 10 elements are defined as:
      ('E_KS', 'E_HF', 'E_GW', 'sigma_x', 'Re_sigma_c', 'Im_sigma_c', 'V_xc', 'delta_HF', 'delta_GW', 'Znk')

    For nitrogen release:
      energies have the shape (n_states, 10), where the 10 elements are defined as:
      ('E_KS', 'E_HF', 'E_GW', 'Sx', 'Sc', 'Vxc', 'DE_HF', 'DE_GW', 'Znk')

    :param str full_file_name: Path + file name
    :return dict data: Parsed k-points and energies from EVALQP.DAT
    """
    try:
        with open(full_file_name) as f:
            file_string = f.read()
    except FileNotFoundError:
        raise FileNotFoundError(f"{full_file_name} does not exist")

    eval_indexing: NumberOfStates = n_states_from_evalqp(file_string)
    k_points = k_points_from_evalqp(file_string)

    # Uses np to parse, hence easier to pass file name here
    energies = parse_evalqp_blocks(full_file_name, k_points, eval_indexing.n_states)
    assert len(k_points) == len(energies), "Should be a set of energies for each k-point"

    data = {'state_range': [eval_indexing.first_state, eval_indexing.last_state],
            'column_labels': parse_column_labels(file_string)}

    # Repackage energies with their respective k-points
    for ik in range(1, len(k_points) + 1):
        data[ik] = {
            'k_point': k_points[ik]['k_point'],
            'weight': k_points[ik]['weight'],
            'energies': energies[ik]
        }
    return data


def parse_column_labels(file_string: str) -> enum.Enum:
    """ Parse the column labels of EVALQP.DAT, which vary between code versions.

    :param str file_string: Input string
    return An enum class with the column labels as attributes, with corresponding
    values starting from 0.
    """
    column_str = file_string.splitlines()[1:2][0]
    column_labels = column_str.split()[1:]
    # zip and count ensure enum indexing starts at 0
    return enum.Enum(value='EvalQPColumns', names=zip(column_labels, count()))


def parse_gw_dos(full_file_name: str) -> dict:
    """Parser for GW DOS files.

    :param full_file_name: Path + file name
    :return dict data: Parsed energies and DOS from GW DOS files
    """
    valid_file_names = ['TDOS.OUT', 'TDOS-QP.OUT']
    path, file_name = os.path.split(full_file_name)
    if file_name not in valid_file_names:
        raise ValueError(f'{file_name} not a valid DOS file name.')

    data = np.genfromtxt(full_file_name)
    return {'energy': data[:, 0], 'dos': data[:, 1]}



""" Band structure class. """

from typing import Tuple, List, Optional, Union
from excitingtools.eigenstates.eigenstates import get_k_point_index
import numpy as np


class BandData:
    ticks_and_labels = Tuple[np.ndarray, List[str]]
    vertex_keys = ['distance', 'label', 'coord']

    def __init__(self,
                 bands: np.ndarray,
                 k_points: np.ndarray,
                 e_fermi: float,
                 flattened_k_points: Optional[np.ndarray] = None,
                 vertices: Optional[List[dict]] = None):
        """ Initialise BandData.

        :param bands: Band energies with shape (n_k_points, n_bands).
        :param: k_points: k-points at which the band energies have been computed.
        :param: e_fermi: Fermi level.
        :param: flattened_k_points: Flattened k-points along which one can plot a band structure.
        :param: vertices: exciting output containing high-symmetry points and symbols along the
        flattened k-path, as defined in exciting.
        """
        self.bands = bands
        self.n_k_points, self.n_bands = self.bands.shape
        self.k_points = k_points
        self.e_fermi = e_fermi
        self.flattened_k_points = flattened_k_points
        self.vertices = vertices
        self.xticks, self.labels = self.band_path()
        self.i_vbm, self.i_cbm = self.get_band_edges()

    def band_path(self) -> ticks_and_labels:
        """ Get an array of points in the k-path that correspond to high symmetry points,
        and a list of their labels.

        vertices expected to have the form
        [{'distance': float, 'label': str, 'coord': [float, float, float]}, ...]
        parsed from exciting's bandstructure.xml file.

        :return: Tuple of NumPy array containing high symmetry points and list containing their labels.
        If vertices is None, return empty containers.
        """
        if self.vertices is None:
            return np.empty(shape=1), []

        assert list(self.vertices[0].keys()) == self.vertex_keys, \
            f'Expect a vertex to have the keys {self.vertex_keys}'

        vertices = [self.vertices[0]["distance"]]
        labels = [self.vertices[0]["label"]]

        for i in range(1, len(self.vertices)):
            vertex = self.vertices[i]["distance"]
            label = self.vertices[i]["label"]

            # Handle discontinuities in the band path
            if np.isclose(vertex, vertices[-1]):
                vertices.pop()
                label = labels.pop() + ',' + label

            vertices.append(vertex)
            labels.append(label)

        # Replace for plotting purposes
        unicode_gamma = '\u0393'
        for label in ['Gamma', 'gamma', 'G']:
            labels = list(map(lambda x: x.replace(label, unicode_gamma), labels))

        return np.asarray(vertices), labels

    def get_k_point_index(self, k_point: Union[List[float], np.ndarray]) -> int:

        if len(k_point) != 3:
            raise TypeError('Expected type for k-point: list or NumPy array of length 3')

        return get_k_point_index(k_point, self.k_points, verbose=False)

    def get_band_edges(self) -> Tuple[int, int]:
        """Get indices of the valence and conduction bands.

        :return Tuple of indices for the valence and conduction bands. A ValueError is returned if the Fermi level is
        located higher than the energy bands.
        """
        valence_max_energies = np.empty(len(self.k_points))

        for k in range(len(self.k_points)):
            valence_max_energies[k] = np.amax(self.bands[k][self.bands[k] <= self.e_fermi])

        vbm_energy = np.amax(valence_max_energies)
        ik_vbm = np.argmax(valence_max_energies)
        n_occupied = len(np.where(self.bands[ik_vbm, :] <= vbm_energy)[0])
        i_vbm = n_occupied - 1

        if i_vbm + 1 >= self.n_bands:
            raise ValueError(f'Fermi level {self.e_fermi} larger than highest band energy {np.amax(self.bands)}')

        return i_vbm, i_vbm + 1

    def get_valence_band_maximum(self) -> float:
        """Get the value of the valence band maximum.
        """
        return np.amax(self.bands[:, self.i_vbm])

    def get_conduction_band_minimum(self) -> float:
        """Get the value of the conduction band minimum.
        """
        return np.amin(self.bands[:, self.i_cbm])

    def get_fundamental_band_gap(self) -> float:
        """Get the value of the fundamental band gap.

        :return fundamental band gap. The band gap is set to zero if bands cross the Fermi level.
        """
        zero_band_gap = 0.0
        if np.any(self.bands[:, self.i_vbm] > self.e_fermi) or np.any(self.bands[:, self.i_cbm] <= self.e_fermi):
            return zero_band_gap

        ik_v = np.argmax(self.bands[:, self.i_vbm])
        ik_c = np.argmin(self.bands[:, self.i_cbm])

        return self.bands[ik_c, self.i_cbm] - self.bands[ik_v, self.i_vbm]

    def get_band_gap(self, k_valence, k_conduction):
        """ Get the value of the band gap calculated between two given k-points.

        :param k_valence: k-point for the valence band.
        :param k_conduction: k-point for the valence band.
        :return band gap.
        """
        ik_v = self.get_k_point_index(k_valence)
        ik_c = self.get_k_point_index(k_conduction)

        return self.bands[ik_c, self.i_cbm] - self.bands[ik_v, self.i_vbm]



"""Functions that operate on a crystal lattice.
"""
import numpy as np

from excitingtools.math.math_utils import triple_product


def check_lattice(lattice: list):
    """Check lattice vector consistency.

    Checks that the user supplies three different lattice vectors.
    One could also check the lattice vector angles.

    :param list lattice: Lattice vectors
    """
    if len(lattice) != 3:
        raise ValueError('lattice argument expected to have 3 elements')

    for i in range(0, 3):
        if len(lattice[i]) != 3:
            raise ValueError(
                f'lattice vector {i} expected to have 3 components. Instead has {len(lattice[i])}'
            )

    for i, j in [(0, 1), (1, 2), (0, 2)]:
        if np.allclose(lattice[i], lattice[j], atol=1.e-6):
            raise ValueError(
                f'lattice vectors {i} and {j} are numerically equivalent')


def check_lattice_vector_norms(lattice: list, tol=1.e-6):
    """ Check the norm of each lattice vector.

    :param list lattice: Lattice vectors
    :param tol: Optional tolerance for what is considered numerically zero.
    """
    norms = np.empty(shape=3)

    for i in range(0, 3):
        norms[i] = np.linalg.norm(lattice[i])

    zero_norms = np.where(norms < tol)[0]
    for i in zero_norms:
        raise ValueError(f'lattice vector {i} has a norm of zero')


def parallelepiped_volume(lattice_vectors: np.ndarray) -> float:
    """Volume of a parallelepiped cell.

    :param np.ndarray lattice_vectors: Lattice vectors, stored column-wise
    :return: float: Cell volume
    """
    return np.abs(
        triple_product(lattice_vectors[:, 0], lattice_vectors[:, 1],
                       lattice_vectors[:, 2]))


def reciprocal_lattice_vectors(a: np.ndarray) -> np.ndarray:
    r"""
    Get the reciprocal lattice vectors of real-space lattice vectors \{\mathbf{a}\}:

      \mathbf{b}_0 = 2 \pi \frac{\mathbf{a}_1 \wedge \mathbf{a}_2} {\mathbf{a}_0 \cdot (\mathbf{a}_1 \wedge \mathbf{a}_2)}
      \mathbf{b}_1 = 2 \pi \frac{\mathbf{a}_2 \wedge \mathbf{a}_3} {\mathbf{a}_0 \cdot (\mathbf{a}_1 \wedge \mathbf{a}_2)}
      \mathbf{b}_2 = 2 \pi \frac{\mathbf{a}_0 \wedge \mathbf{a}_1} {\mathbf{a}_0 \cdot (\mathbf{a}_1 \wedge \mathbf{a}_2)}

    :param np.ndarray a: Lattice vectors, stored column-wise
    :return: np.ndarray b: Reciprocal lattice vectors, stored column-wise
    """
    volume = triple_product(a[:, 0], a[:, 1], a[:, 2])
    b = np.empty(shape=(3, 3))
    b[:, 0] = 2 * np.pi * np.cross(a[:, 1], a[:, 2]) / volume
    b[:, 1] = 2 * np.pi * np.cross(a[:, 2], a[:, 0]) / volume
    b[:, 2] = 2 * np.pi * np.cross(a[:, 0], a[:, 1]) / volume
    return b


# TODO(Bene) This needs cleaning up. Missing any documenting maths. Not at all clear what's happening
def plane_transformation(rec_lat_vec: np.array,
                         plot_vec: np.array) -> np.array:
    """
    Take reciprocal lattice vectors and ONS of a plane in rec. lat. coordinates where the first two vectors span the plane and the third is normal to them
    and calculate a matrix that transforms points in the plane to the xy plane in cartesian coordinates.
    input:
    :param rec_lat_vec:            reciprocal lattice vectors
    :param plot_vec:               ONS of the plotting plane
    :return transformation_matrix: matrix that transforms k and spin vectors to the plot plane
    """
    norm = np.linalg.norm
    # transform plot vec in cartesian coordinates
    plot_vec = (rec_lat_vec.dot(plot_vec)).transpose()
    # extend plot vec to an orthogonal system
    plot_vec = np.array([(plot_vec[1] - plot_vec[0]) / norm(plot_vec[1] - plot_vec[0]),
                         (plot_vec[2] - plot_vec[0]) / norm(plot_vec[2] - plot_vec[0]),
                         np.cross(plot_vec[1] - plot_vec[0], plot_vec[2] - plot_vec[0]) \
                         / norm(np.cross(plot_vec[1] - plot_vec[0], plot_vec[2] - plot_vec[0]))])
    transformation_matrix = np.linalg.inv(plot_vec)
    for v in transformation_matrix:
        v = v / norm(v)
    transformation_matrix = np.transpose(transformation_matrix)

    return transformation_matrix



"""General utility functions. Typically, conversion/type-checking.
"""
import pathlib
import re
from typing import Union, List, Optional, Callable, Iterator


def get_excitingtools_root() -> pathlib.Path:
    """ Get the root directory of excitingtools. """
    return pathlib.Path(__file__).parents[2]


def can_be_float(value) -> bool:
    """
    Check if a value can be interpreted as a float

    :param value: Input
    :return bool: Value can be interpreted as a float
    """
    try:
        float(value)
        return True
    except ValueError:
        return False


def convert_to_literal(input_str: str) -> Union[int, float, None]:
    """
    If possible, convert string to an int or float

    For example:
      convert_to_literal('1.1') returns 1.1
      convert_to_literal('1') returns 1
      convert_to_literal('1.0') returns 1.0

    :param str input_str: Input string
    :return Union[int, float] x: Numerical literal of x, else None.
    """
    try:
        integer_string = int(input_str)
        return integer_string
    except ValueError:
        try:
            float_string = float(input_str)
            return float_string
        except ValueError:
            return None


def get_new_line_indices(string: str) -> List[int]:
    """
    Given a string, return the indices that correspond to the
    start of new lines.

    For example,
     line = get_new_line_indices(string)
     # First line
     string[line[0]: line[1]]
     # 6th line
     string[line[5]: line[6]]

    :param str string: Input string
    :return List[int] indices: List of indices corresponding to
     new lines in string.
    """
    indices = [0]
    indices += [m.start() + 1 for m in re.finditer('\n', string)]
    return indices


def list_to_str(mylist: list, modifier: Optional[Callable] = None) -> str:
    """ Convert a list to a string
    """
    if modifier is None:
        modifier = lambda x: x
    return "".join(modifier(str(xyz)) + ' ' for xyz in mylist).strip()


def flatten_list(input_list: list) -> Iterator:
    """ Flatten a list of lists and other elements.

    :param input_list: input list
    :return: an iterator for the flattened list
    """
    for x in input_list:
        if isinstance(x, list):
            yield from flatten_list(x)
        else:
            yield x



"""SpeciesFile class.
Provides functionality to read in a species file, add high energy local orbitals (HELOs), and 
write the updated XML structures back to file.
"""
import copy
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Union
from xml.etree import ElementTree
from collections import defaultdict 

from excitingtools.exciting_dict_parsers.species_parser import parse_species_xml
from excitingtools.input.xml_utils import xml_tree_to_pretty_str
from excitingtools.utils.jobflow_utils import special_serialization_attrs


@dataclass
class SpeciesFile:
    """ Holds a species file. """
    species: dict
    muffin_tin: dict
    atomic_states: list
    basis: dict 

    def __post_init__(self):
        """Ensures that the 'basis' dictionary includes 'custom' and 'lo' keys. 
        If these keys are absent, they are initialized as empty lists. 
        """
        self.basis.setdefault('custom', [])
        self.basis.setdefault('lo', [])

    @classmethod
    def from_file(cls, file: Union[str, Path]):
        """ Reads in an exciting species XML file, parses the file as a dict and
        initializes an instance of the class with the parsed data.
        
        :param file: XML species file.
        :return: An instance of the class initialized with the data from the species file.
        """
        return cls(**parse_species_xml(file))
    
    def as_dict(self) -> dict:
        """Returns a dictionary representing the current object for later recreation.
        The serialise attributes are required for recognition by monty and jobflow.
        """
        serialise_attrs = special_serialization_attrs(self)
        return {**serialise_attrs, **self.__dict__}

    @classmethod
    def from_dict(cls, d: dict):
        my_dict = copy.deepcopy(d)
        # Remove key value pairs needed for workflow programs
        # call function on class to get only the keys (values not needed)
        serialise_keys = special_serialization_attrs(cls)
        for key in serialise_keys:
            my_dict.pop(key, None)
        return cls(**my_dict)

    def get_first_helo_n(self, l: int) -> int:
        """Returns the first principle quantum number 'n' for which additional High Energy Local
        Orbitals (HELOs) can be added, for a given angular momentum quantum number 'l'. 
        The 'n' value determination is based on:

        1. For a specific 'l' channel the highest 'n' specified in the atomic states is 'max_n'.
        2. If there exist local orbitals with a 'n' greater than the maximum found in atomic states for 
           the same 'l', 'max_n' is set to this 'n'.
        3. The first added HELO then starts at 'max_n + 1'.
        4. For an 'l' channel not represented in atomic states or as a local orbital, the first added HELO 
            has principle quantum number 'n = l + 1' .

        :param l: angular momentum quantum number 'l' 
        :return: first possible principle quantum number 'n' for a HELO for the given 'l'-channel 
        """
        atomicstate_ns_per_l, lo_ns_per_l = self.get_atomicstate_ns_per_l(), self.get_lo_ns_per_l()
        max_n_for_specified_l = max(atomicstate_ns_per_l.get(l, [l]) + lo_ns_per_l.get(l, [l]))

        # If matchingOrder > 1 for the same l and highest n, skip n+1 HELO.
        ns_per_l_with_mO_over_1 = self.check_matching_orders()
        if ns_per_l_with_mO_over_1.get(l) and max_n_for_specified_l in ns_per_l_with_mO_over_1[l]:
            warnings.warn(f"Warning: HELO skipped for l: {l} and n: {max_n_for_specified_l + 1}")
            return max_n_for_specified_l + 2

        return max_n_for_specified_l + 1
    
    def get_atomicstate_ns_per_l(self) -> Dict[int, list]:
        """Generates a dictionary mapping each 'l' channel present in the atomic states their corresponding 'n' values. 
        Only includes local orbitals with a defined 'n' value. Local orbitals only defined through the attribute 
        'trialEnergy' are ignored.

        :return: Dictionary with 'l' channels as keys and lists of 'n' values from the atomic states.
        """
        atomicstate_ns_per_l = defaultdict(list)
        for state in self.atomic_states:
            atomicstate_ns_per_l[state["l"]].append(state["n"])

        return atomicstate_ns_per_l

    def get_lo_ns_per_l(self) -> Dict[int, list]:
        """Generates a dictionary mapping each 'l' channel present in the local orbitals to their corresponding 'n' 
        values. Only includes local orbitals with a defined 'n' value. Local orbitals only defined through the 
        attribute 'trialEnergy' are ignored.

        :return: Dictionary with 'l' channels as keys and corresponding lists of 'n' values from local orbitals.
        """
        lo_ns_per_l = defaultdict(list)
        for lo in self.basis["lo"]:
            for wf in lo["wf"]:
                if "n" in wf:
                    lo_ns_per_l[lo["l"]].append(wf["n"])

        return lo_ns_per_l
    
    def get_helos_from_species(self) -> Dict[int, set]:
        """Generates a dictionary mapping each 'l' channel to a list of principle quantum numbers 'n' unique to high
        energy/local orbitals (HELOs). This is achieved by comparing 'n' values in local orbitals against those
        in atomic states for each 'l' channel, extracting those exclusive to HELOs.

        :return: Dictionary with 'l' channels as keys and exclusive HELO 'n' values as sets.
        """
        atomicstate_ns_per_l, lo_ns_per_l = self.get_atomicstate_ns_per_l(), self.get_lo_ns_per_l()
        
        helo_ns_per_l = {}
        for l, ns in lo_ns_per_l.items():
            unique_helos = set(ns) - set(atomicstate_ns_per_l.get(l, []))
            helo_ns_per_l[l] = unique_helos

        return helo_ns_per_l

    def check_matching_orders(self) -> Dict[int, list]:
        """Creates a dictionary that lists principle quantum numbers 'n' for each angular momentum quantum number 'l', 
        where the 'n' values are associated with a matchingOrder greater than 1 in the local orbitals.

        :return: Dictionary with 'l' channels as keys and lists of 'n' values that have matchingOrder > 1 as values.
        """
        ns_per_l_with_mO_over_1 = defaultdict(list)
        for lo in self.basis["lo"]:
            for wf in lo["wf"]:
                if wf.get("n") and wf["matchingOrder"] > 1:
                    ns_per_l_with_mO_over_1[lo["l"]].append(wf["n"])

        return ns_per_l_with_mO_over_1

    def add_helos(self, l: int, number: int):
        """Adds a specific 'number' of High Energy Local Orbitals (HELOs) to the basis 
        for a given angular momentum channel 'l'.

        :param l: angular momentum number l
        :param number: the number of HELOs to be added to the l-channel
        """
        first_helo_n_for_l = self.get_first_helo_n(l)

        for nr_lo in range(number):
            new_lo = {
                'l': l,
                'wf': [
                    {'matchingOrder': 0, 'searchE': 'False', 'n': first_helo_n_for_l + nr_lo},
                    {'matchingOrder': 1, 'searchE': 'False', 'n': first_helo_n_for_l + nr_lo}
                ]
            }
            self.basis['lo'].append(new_lo)

    def to_xml(self) -> ElementTree.Element:
        """Converts the class attributes into an XML structure using ElementTree.

        :return: An ElementTree.Element representing the root of the XML structure.
        """
        spdb = ElementTree.Element("spdb")
        sp = ElementTree.SubElement(spdb, "sp", {k: str(v) for k, v in self.species.items()})
        ElementTree.SubElement(sp, "muffinTin", {k: str(v) for k, v in self.muffin_tin.items()})
        for state in self.atomic_states:
            ElementTree.SubElement(
                sp, "atomicState", {k: str(v).lower() for k, v in state.items()}) 

        basis = ElementTree.SubElement(sp, "basis")
        for default in self.basis["default"]:
            ElementTree.SubElement(
                basis, "default", {k: str(v).lower() for k, v in default.items()})

        for custom in self.basis["custom"]:
            ElementTree.SubElement(
                basis, "custom", {k: str(v).lower() for k, v in custom.items()})

        for lo in self.basis["lo"]:
            lo_element = ElementTree.SubElement(basis, "lo", l=str(lo["l"]))
            for wf in lo['wf']:
                ElementTree.SubElement(
                    lo_element, "wf", {k: str(v).lower() for k, v in wf.items()})

        return spdb

    def to_xml_str(self) -> str:
        """Compose XML ElementTrees from exciting input classes to create an input xml string.

        :return: Input XML tree as a string, with pretty formatting.
        """
        return xml_tree_to_pretty_str(self.to_xml())

    def write(self, filename: Union[str, Path]):
        """Writes the xml string to file.

        :param filename: name of the file.
        """
        with open(filename, "w") as fid:
            fid.write(self.to_xml_str())



""" Test the utilities for ase. """
from ase.build import bulk

from excitingtools import ExcitingStructure
from excitingtools.structure.ase_utilities import exciting_structure_to_ase


def test_class_exciting_structure_to_ase():
    ase_atoms = bulk("Si")
    structure = ExcitingStructure(ase_atoms, species_path='./')
    new_ase_atoms = exciting_structure_to_ase(structure)
    assert ase_atoms.wrap() == new_ase_atoms.wrap()



"""Wrapper for command-line grep.
"""
import subprocess
from typing import Optional, Union


def grep(string: str, fname: str, options: Optional[dict] = None) -> Union[str, None]:
    """
    Wrapper for command-line grep.

    Can pass any grep options like: options = {'A': value}

    :param str string: Search string
    :param str fname: File name to search
    :param Optional[dict] options: Grep options.
    :return output: String if matched, None if failed.
    """
    opts = ''
    if options is not None:
        for key, value in options.items():
            opts += '-' + key + ' ' + str(value) + ' '

    grep_str = "grep " + opts + " '" + string + "' " + fname

    try:
        output = subprocess.check_output(grep_str, shell=True).decode("utf-8")
    except subprocess.CalledProcessError as grepexc:
        print("subprocess error:", grepexc.returncode, "grep found:",
              grepexc.output)
        output = None

    return output



"""
Test for the input.xml file parser
"""

from excitingtools.exciting_obj_parsers.input_xml import parse_input_xml

reference_input_str = """<?xml version="1.0" encoding="UTF-8"?>
<input>
  
  <title>Lithium Fluoride BSE</title>
  
  <structure speciespath="." autormt="false" epslat="1.0d-6">
    <crystal scale="1.0" stretch="1.0">
      <basevect>3.80402 3.80402 0.00000</basevect>
      <basevect>3.80402 0.00000 3.80402</basevect>
      <basevect>0.00000 3.80402 3.80402</basevect>
    </crystal>
    <species speciesfile="Li.xml" rmt="1.5">
      <atom coord="0.0000  0.0000  0.0000" bfcmt="0.0 0.0 0.0"/>
    </species>
    <species speciesfile="F.xml">
      <atom coord="0.5000  0.5000  0.5000" lockxyz="false true false"/>
    </species>
  </structure>
  
  <groundstate xctype="GGA_PBE" ngridk="4 4 4" epsengy="1d-7" outputlevel="high"/>

  <xs xstype="BSE" 
   ngridq="3 3 3"
   vkloff="0.05 0.15 0.25" 
   nempty="1"
   broad="0.0073499"
   nosym="true">

   <energywindow intv="0.0 1.0" 
    points="50" />

   <screening screentype="full"
    nempty="115" />

   <BSE bsetype="singlet"
    nstlbse="1 5 1 2" 
    aresbse="false"/>

   <qpointset>
      <qpoint>0.0 0.0 0.0</qpoint>
   </qpointset>
   
   <plan>
      <doonly task="screen" />
      <doonly task="bse" />
   </plan>
  </xs>
  
</input>
"""


def test_parse_input_xml_to_object():
    input_xml = parse_input_xml(reference_input_str)
    assert set(vars(input_xml)) == {'xs', 'groundstate', 'structure', 'title'}
    assert input_xml.to_xml_str().startswith('<?xml version="1.0" ?>\n<input>\n\t<title>Lithium Fluoride BSE')



""" Tests for pymatgen utilities. """
import numpy as np
import pytest


@pytest.fixture
def pymatgen_atoms_H2O():
    """
    H20 molecule in a big box (angstrom), in pymatgen Structure()
    Converts a List[dict] to pymatgen.core.structure.Structure.
    """
    pymatgen_struct = pytest.importorskip("pymatgen.core.structure")
    cubic_cell = np.array([[10.0, 0.0, 0.0], [0.0, 10.0, 0.0], [0.0, 0.0, 10.0]])
    atoms = [{'species': 'H', 'position': [0.00000, 0.75545, -0.47116]},
             {'species': 'O', 'position': [0.00000, 0.00000, 0.11779]},
             {'species': 'H', 'position': [0.00000, 0.75545, -0.47116]}]

    symbols = [atom['species'] for atom in atoms]
    positions = [atom['position'] for atom in atoms]
    return pymatgen_struct.Structure(lattice=cubic_cell, species=symbols, coords=positions, coords_are_cartesian=True)


def test_class_exciting_structure_pymatgen(pymatgen_atoms_H2O):
    """
    Test the pymatgen Structure object gets used correctly by the ExcitingStructure constructor.
    """
    pymatgen_conversion = pytest.importorskip("excitingtools.structure.pymatgen_utilities")
    structure = pymatgen_conversion.pymatgen_to_exciting_structure(pymatgen_atoms_H2O)

    assert structure.species == ["H", "O", "H"]
    assert np.allclose(structure.lattice,
                       [[18.897261246257703, 0.0, 0.0],
                        [0.0, 18.897261246257703, 0.0],
                        [0.0, 0.0, 18.897261246257703]]), \
        'Expect lattice vectors to match input values'

    assert np.allclose(structure.positions, pymatgen_atoms_H2O.frac_coords), 'Expect positions to match input values.'

    # This just confirms the XML tree is built, not that it is correct.
    xml_structure = structure.to_xml()
    assert list(xml_structure.keys()) == ['speciespath'], 'Only expect speciespath in structure xml keys'


def test_class_exciting_structure_to_pymatgen(pymatgen_atoms_H2O):
    pymatgen_conversion = pytest.importorskip("excitingtools.structure.pymatgen_utilities")
    structure = pymatgen_conversion.pymatgen_to_exciting_structure(pymatgen_atoms_H2O)
    new_pymatgen_atoms = pymatgen_conversion.exciting_structure_to_pymatgen(structure)
    assert pymatgen_atoms_H2O == new_pymatgen_atoms



"""Parsers for GW's EPS00_GW.OUT output.
"""
import numpy as np

from excitingtools.parser_utils.parser_decorators import accept_file_name
from excitingtools.utils.utils import get_new_line_indices

# Output file name
_file_name = 'EPS00_GW.OUT'


def parse_eps00_frequencies(file_string: str) -> dict:
    """Parse frequencies from EPS00_GW.OUT.

    :param str file_string: Input string
    :return dict frequencies: Dictionary of frequencues {index: frequency}
    """
    initial_header_size = 3
    block_size = 6

    # Start file_lines on first block header
    file_lines = file_string.splitlines()
    file = file_lines[initial_header_size:]
    n_freq = int(len(file) / block_size)

    frequencies = {}
    j = 0
    for i in range(0, n_freq):
        j = i * block_size
        frequencies[i + 1] = float(file[j].split()[-1])

    assert len(frequencies) == n_freq

    return frequencies


def parse_eps00_blocks(file_string: str, n_freq: int) -> dict:
    """Parser for epsilon blocks in EPS00_GW.OUT.

    File of the form:

    (dielectric tensor, random phase approximation)

    frequency index and value:      1    0.01985507
    real part, imaginary part below
     152.69882148    0.00000000    0.00000000         0.00000000    0.00000000    0.00000000
       0.00000000  152.69882148    0.00000000         0.00000000    0.00000000    0.00000000
       0.00000000    0.00000000  152.69882148         0.00000000    0.00000000    0.00000000

    frequency index and value:      2    0.02771249
     real part, imaginary part below
        8.22189228    0.00000000    0.00000000        -0.00000000    0.00000000    0.00000000
        0.00000000    8.22189228    0.00000000         0.00000000   -0.00000000    0.00000000
        0.00000000    0.00000000    8.22189228         0.00000000    0.00000000   -0.00000000

    :param str file_string: Input string
    :param int n_freq: Number of frequency points
    :return dict data: Dict containing real and imaginary parts of the dielectric function,
     eps00, for each frequency point. Frequency indexing (keys) start at 1.
    """
    line = get_new_line_indices(file_string)
    assert file_string[line[0]:line[1]].isspace(), (
        "First line of EPS00_GW.OUT must be a whiteline")

    initial_header = 3
    offset = initial_header - 1
    # Header lines + eps00
    repeat_block_size = 6

    data = {}
    file_string = file_string.splitlines()[initial_header:]

    def extract_eps_i(line: str) -> tuple:
        eps_i_re = np.array(line.split()[0:3], dtype=float)
        eps_i_img = np.array(line.split()[3:6], dtype=float)
        return eps_i_re, eps_i_img

    for i_freq in range(0, n_freq):
        i = offset + i_freq * repeat_block_size

        eps_x_re, eps_x_img = extract_eps_i(file_string[i])
        eps_y_re, eps_y_img = extract_eps_i(file_string[i + 1])
        eps_z_re, eps_z_img = extract_eps_i(file_string[i + 2])

        data[i_freq + 1] = {
            're': np.array([eps_x_re, eps_y_re, eps_z_re]),
            'img': np.array([eps_x_img, eps_y_img, eps_z_img])
        }

    return data


@accept_file_name
def parse_eps00_gw(file_string: str) -> dict:
    """ Parser frequency grid and epsilon00 from EPS00_GW.OUT.

    :param str file_string: Input string
    :return dict data: Dict containing the frequency, and real and imaginary parts
     of the dielectric function, eps00, for each frequency point.
    """
    frequencies = parse_eps00_frequencies(file_string)
    eps00 = parse_eps00_blocks(file_string, len(frequencies))
    assert len(frequencies) == len(eps00), \
        "Expect eps00 to have n frequencies consistent with the frequency grid"

    # Repackage frequency points and eps00 together
    data = {}
    for i_freq in range(1, len(frequencies) + 1):
        data[i_freq] = {'frequency': frequencies[i_freq], 'eps00': eps00[i_freq]}

    return data



"""Structure class, mirroring that of exciting's structure XML sub-tree.
https://exciting.wikidot.com/ref:structure
"""
from __future__ import annotations

import copy
from pathlib import Path
from typing import Optional, Union, List, Dict
from xml.etree import ElementTree

import numpy as np

from excitingtools.constants.units import angstrom_to_bohr, bohr_to_angstrom
from excitingtools.input.base_class import ExcitingXMLInput, AbstractExcitingInput
from excitingtools.input.input_classes import ExcitingCrystalInput, ExcitingSpeciesInput
from excitingtools.structure.lattice import check_lattice, check_lattice_vector_norms
from excitingtools.utils import valid_attributes
from excitingtools.utils.utils import list_to_str


class ExcitingStructure(ExcitingXMLInput):
    """ Class allowing exciting XML structure to be written from python data.
    """
    name = "structure"
    # Path type
    path_type = Union[str, Path]

    # Mandatory attribute "coord" taken out because it's specified inside the atoms
    _valid_atom_attributes = set(valid_attributes.atom_valid_attributes) - {"coord"}

    def __init__(self,
                 atoms,
                 lattice: Optional[list | np.ndarray] = None,
                 species_path: path_type = './',
                 crystal_properties: Optional[dict | ExcitingCrystalInput] = None,
                 species_properties: Optional[Dict[str, Union[dict, ExcitingSpeciesInput]]] = None,
                 **kwargs):
        """ Initialise instance of ExcitingStructure.

        TODO(Alex) Issue 117. Create our own class with a subset of methods common to ASE' Atom()
          Then we can have a single API for this init. If ASE is used, xAtom() is just a wrapper of
          Atom(), else we have some light methods.

        All valid attributes can be found in the module valid_attributes.py

        :param atoms: Atoms object of type ase.atoms.Atoms or of the form List[dict], for example:
         atoms = [{'species': 'X', 'position': [x, y, z]}, ...].
         Each dict can also optionally contain the _valid_atom_attributes:
         {'species': 'X', 'position': [x, y, z],
           'bfcmt': [bx, by, bz], 'lockxyz': [lx, ly, lz], 'mommtfix': [mx, my, mz]}.
        If atoms are defined with ASE, optional atomic_properties cannot be specified.
        Eventually, the list of atoms will be replaced with our custom class, which will extend ase.Atoms()
        with the additional, optional attributes.
        Species value can be a file_name without the suffix '.xml', which will be added automatically.

        :param lattice [a, b, c], where a, b and c are lattice vectors with 3 components.
         For example, a = [ax, ay, az]. Only required if one does not pass an ase Atoms object.
        :param species_path: Optional path to the location of species file/s.
        :param crystal_properties: Optional crystal properties.
        :param species_properties: Optional species properties, defined as:
        {'species1': {'rmt': rmt_value}, 'species2': {'rmt': rmt_value}}
        and with subtrees as:
        {'species1': {'rmt': rmt_value, 'LDAplusU': {'J': J, 'U': U, 'l': l}}, species2: ... }
        :param kwargs: Optional structure properties. Passed as kwargs.
        """
        if isinstance(species_path, Path):
            species_path = species_path.as_posix()
        super().__init__(speciespath=species_path, **kwargs)
        self.structure_attributes = {"speciespath": species_path, **kwargs}

        if isinstance(atoms, list) and lattice is None:
            raise ValueError("If atoms is a list, lattice must be passed as a separate argument.")

        # Simple container for atoms, as documented in __init__.
        if isinstance(atoms, list):
            check_lattice(lattice)
            check_lattice_vector_norms(lattice)
            self.lattice = np.asarray(lattice)
            self.species = [atom['species'].capitalize() for atom in atoms]
            self.positions = [atom['position'] for atom in atoms]
            self.atom_properties = self._init_atom_properties(atoms)
        else:
            self.lattice, self.species, self.positions = self._init_lattice_species_positions_from_ase_atoms(atoms)
            self.atom_properties = [{}] * len(self.species)

        self.unique_species = sorted(set(self.species))

        # Optional properties
        self.crystal_properties = self._initialise_subelement_attribute(ExcitingCrystalInput,
                                                                        crystal_properties or {})
        self.species_properties = dict(self._init_species_properties(species_properties))

    def __setattr__(self, name: str, value):
        """ Overload the attribute setting from the base class, since here we use different attribute names than
        defined in the schema.

        :param name: name of the attribute
        :param value: new value, can be anything
        """
        AbstractExcitingInput.__setattr__(self, name, value)

    def _init_lattice_species_positions_from_ase_atoms(self, atoms) -> tuple:
        """ Initialise lattice, species and positions from an ASE Atoms Object.

        Duck typing for atoms, such that ASE is not a hard dependency.

        :param atoms: ASE Atoms object.
        :return  Lattice, species, positions: Lattice, species and positions
        """
        try:
            cell = atoms.get_cell()
            # ASE works in Angstrom, whereas exciting expects atomic units
            lattice = np.asarray(cell) * angstrom_to_bohr
            species = [x.capitalize() for x in atoms.get_chemical_symbols()]
            if self.structure_attributes.get("cartesian"):
                positions = angstrom_to_bohr * atoms.get_positions()
            else:
                positions = atoms.get_scaled_positions()
            return lattice, species, positions
        except AttributeError:
            message = "atoms must either be an ase.atoms.Atoms object or List[dict], of the form" \
                      "[{'species': 'X', 'position': [x, y, z]}, ...]."
            raise AttributeError(message)

    def _init_atom_properties(self, atoms: List[dict]) -> List[dict]:
        """ Initialise atom_properties.

        For atoms that contain optional atomic properties, store them as
        dicts in a list of len(n_atoms). Atoms with none of these properties
        will be defined as empty dicts.

        For each element of atoms, one must have  {'species': 'X', 'position': [x, y, z]}  and
        may have the additional attributes: {'bfcmt': [bx, by, bz], 'lockxyz': [lx, ly, lz], 'mommtfix': [mx, my, mz]}.
        Extract the optional attributes and return in `atom_properties`, with string values.

        :param atoms: List container.
        :return atom_properties: List of atom properties. Each element is a dict.
        and the dict value has been converted to string - ready for XML usage.
        """
        atom_properties: List[dict] = []

        for atom in atoms:
            optional_property_keys = set(atom.keys()) & self._valid_atom_attributes
            optional_atom = {key: atom[key] for key in optional_property_keys}
            optional_properties = {}
            for key, value in optional_atom.items():
                optional_properties[key] = self._attributes_to_input_str[type(value)](value)
            atom_properties.append(optional_properties)

        return atom_properties

    def _init_species_properties(self, species_properties: dict) -> Dict[str, ExcitingXMLInput]:
        """ Initialise species_properties.

        For species without properties, return empty_properties: {'S': {}, 'Al': {}}.

        :param species_properties: Species properties
        :return Dict of ExitingXMLInput-species_properties.
        """
        if species_properties is None:
            species_properties = {}

        for species in self.unique_species:
            props = species_properties.get(species) or {}
            props["speciesfile"] = species + '.xml'
            yield species, self._initialise_subelement_attribute(ExcitingSpeciesInput, props)

    def get_lattice(self, convert_to_angstrom: bool = False) -> np.ndarray:
        """ Get the full lattice, meaning after the application of scale and stretch values to the stored
        lattice vectors.

        :param convert_to_angstrom: if True returns lattice in angstrom, else in bohr
        :return: full lattice vectors, stored row-wise in a matrix
        """
        lattice = copy.deepcopy(self.lattice)
        unit_conversion = bohr_to_angstrom if convert_to_angstrom else 1
        scale = getattr(self.crystal_properties, "scale", 1)
        lattice *= scale * unit_conversion

        stretch = np.array(getattr(self.crystal_properties, "stretch", [1, 1, 1]))
        lattice *= stretch[:, None]
        return lattice

    def _group_atoms_by_species(self) -> dict:
        """Get the atomic indices for atoms of each species.

        For example, for:
          species = ['Cd', 'S', 'Cd]
        return:
          indices = {'Cd': [1, 3], 'S' : [2]}

        :return dict indices: Indices of atoms in species and positions
        """
        indices = {}
        for x in self.unique_species:
            indices[x] = [i for i, element in enumerate(self.species) if element == x]
        return indices

    def _xml_atomic_subtree(self, x: str, species, atomic_indices):
        """ Add the required atomic positions and any optional attributes, per species.

        :param x: Species
        :param species: Empty SubElement for species x, which gets filled
        """
        for iatom in atomic_indices[x]:
            coord_str = list_to_str(self.positions[iatom])
            ElementTree.SubElement(species, "atom", coord=coord_str, **self.atom_properties[iatom]).text = ' '

    def to_xml(self) -> ElementTree.Element:
        """Convert structure attributes to XML ElementTree
        Makes use of the to_xml() function of the ExitingXMLInput class to convert values to string.

        Expect to return an XML structure which looks like:
          <structure speciespath="./">

           <crystal scale="1.00" scale="1.00" >
             <basevect>1.0 1.0 0.0</basevect>
             <basevect>1.0 0.0 1.0</basevect>
             <basevect>0.0 1.0 1.0</basevect>
           </crystal>

           <species speciesfile="Al.xml">
             <atom coord="0.0  0.0  0.0"> </atom>
           </species>

          </structure>

        :return ET structure: Element tree containing structure attributes.
        """
        structure_attributes = {
            key: self._attributes_to_input_str[type(value)](value) for key, value in self.structure_attributes.items()
        }
        structure = ElementTree.Element(self.name, **structure_attributes)
        structure.text = ' '

        # Lattice vectors
        crystal = self.crystal_properties.to_xml()
        structure.append(crystal)
        for vector in self.lattice:
            ElementTree.SubElement(crystal, "basevect").text = list_to_str(vector)

        # Species tags
        atomic_indices = self._group_atoms_by_species()
        for x in self.unique_species:
            species = self.species_properties[x].to_xml()
            structure.append(species)
            self._xml_atomic_subtree(x, species, atomic_indices)

        return structure



"""Wrappers for parsing with regex
"""
import re
from typing import List, Union

from excitingtools.utils.utils import convert_to_literal


def parse_value_regex(file_string: str,
                      key: str,
                      no_colon=True,
                      silent_key_error=False) -> dict:
    """
    Match the first instance of a string (key) if present in file_string,
    and return the result in a dictionary.

    :param str file_string: Input string
    :param str key: String to match, also used as a key in the returned dictionary
    :param optional bool no_colon: Remove trailing colons from parsed data keys
    :param optional bool silent_key_error: Print key error

    :return dict data: Matched data
    """
    data = {}

    def process_value(x: str) -> Union[int, float, str]:
        numerical_value = convert_to_literal(x)
        if numerical_value is None:
            return x.strip()
        else:
            return numerical_value

    try:
        match = re.search(key + '(.+)\n', file_string)
        values = match.group(1).split()
        # Remove escape characters
        parser_key = key.replace('\\', "")
        processed_values = [process_value(raw_value) for raw_value in values]
        data[parser_key] = processed_values[0] if len(
            processed_values) == 1 else processed_values

    except AttributeError:
        if not silent_key_error:
            print("parse_value_regex. Did not find the key:", key)
        return {}

    if no_colon:
        return {key.rstrip(':'): value for key, value in data.items()}
    else:
        return data


def parse_values_regex(file_string: str, keys: List[str]) -> dict:
    """
    For a list of strings, match the first instance of each string
    (contained in keys) if present in file_string, and return the
    result in a dictionary.

    :param str file_string: Input string
    :param List[str] keys: Keys to match in lines of file_string

    :return dict parsed_data: Matched data
    """
    assert type(keys) == list, "2nd argument of parse_values_regex must be List[str]"
    matches_dict = {}
    for key in keys:
        matches_dict.update(**parse_value_regex(file_string, key))
    return matches_dict



""" Generate bandstructure input from atoms. """
from __future__ import annotations

import re
from typing import List

import ase.dft
import numpy as np
from ase import Atoms
from ase.cell import Cell

from excitingtools import ExcitingStructure
from excitingtools.input.input_classes import ExcitingBandStructureInput

_default_steps = 100


def band_structure_input_from_cell_or_bandpath(
        cell_or_bandpath: Cell | ase.dft.kpoints.BandPath,
        steps: int = _default_steps) -> ExcitingBandStructureInput:
    """Get band path from ASE lattice cell or ASE bandpath object.

    :param cell_or_bandpath: ase.cell.Cell object or ase.dft.kpoints.BandPath
    :param steps: number of steps for bandstructure calculation
    :return: the exciting band structure input
    """
    bandpath = cell_or_bandpath.bandpath() if isinstance(cell_or_bandpath, Cell) else cell_or_bandpath
    points = []
    pattern = r'[A-Z,][a-z]*[0-9]*'
    for point in re.findall(pattern, bandpath.path):
        if point == ",":
            points[-1]["breakafter"] = True
        else:
            points.append(
                {"coord": list(bandpath.special_points[point]), "label": point})

    return ExcitingBandStructureInput(plot1d={"path": {"steps": steps, "point": points}})


def band_structure_input_from_lattice(lattice_vectors: List[List[float]] | np.ndarray,
                                      steps: int = _default_steps) -> ExcitingBandStructureInput:
    """ Get band path from lattice vectors as array or list of lists using ASE. Lattice vectors
    in array needs to be stored row-wise.

    :param lattice_vectors: lattice
    :param steps: number of steps for bandstructure calculation
    :return: the exciting band structure input
    """
    return band_structure_input_from_cell_or_bandpath(Cell(lattice_vectors), steps=steps)


def band_structure_input_from_ase_atoms_obj(atoms_obj: Atoms,
                                            steps: int = _default_steps) -> ExcitingBandStructureInput:
    """ Get band path from ase object.

    :param atoms_obj: ase atoms object
    :param steps: number of steps for bandstructure calculation
    :return: the exciting band structure input
    """
    return band_structure_input_from_cell_or_bandpath(atoms_obj.cell, steps)


def get_bandstructure_input_from_exciting_structure(structure: ExcitingStructure,
                                                    steps: int = _default_steps) -> ExcitingBandStructureInput:
    """ Use ase bandpath to get bandstructure input for exciting using ASE.

    :param structure: exciting structure input object
    :param steps: number of steps for bandstructure calculation
    :return: bandstructure input for exciting
    """
    return band_structure_input_from_lattice(structure.get_lattice(), steps=steps)



"""Parsers for BSE output files.
"""
import re
from typing import Optional

import numpy as np


def numpy_gen_from_txt(name: str, skip_header: Optional[int] = 0) -> np.ndarray:
    """  Numpy genfromtxt, dressed in try/expect.

    Not worth generalising, as would need to support genfromtxt's API.

    :param name: File name.
    :param skip_header: Optional number of header lines to skip.
    :return data: Parsed data.
    """
    try:
        data = np.genfromtxt(name, skip_header=skip_header)
    except ValueError:
        raise ValueError(f'Failed to parse {name}')
    return data


def parse_EPSILON_NAR(name: str) -> dict:
    """
    Parser for:
        EPSILON_NAR_BSE-singlet-TDA-BAR_SCR-full_OC.OUT.xml,
        EPSILON_NAR_FXCMB1_OC_QMT001.OUT.xml,
        EPSILON_NAR_NLF_FXCMB1_OC_QMT001.OUT.xml,
        LOSS_NAR_FXCMB1_OC_QMT001.OUT.xml
    """
    data = numpy_gen_from_txt(name, skip_header=14)
    out = {
        "frequency": data[:, 0],
        "real_oscillator_strength": data[:, 1],
        "imag_oscillator_strength": data[:, 2],
        "real_oscillator_strength_kkt": data[:, 3]
    }
    return out


def parse_LOSS_NAR(name):
    """
    Parser for:
     LOSS_NAR_FXCMB1_OC_QMT001.OUT.xml,
     LOSS_NAR_NLF_FXCMB1_OC_QMT001.OUT.xml
    """
    data = numpy_gen_from_txt(name, skip_header=14)
    out = {
        "frequency": data[:, 0],
        "real_oscillator_strength": data[:, 1],
        "imag_oscillator_strength": data[:, 2]
    }

    return out


def parse_EXCITON_NAR_BSE(name):
    """
    Parser for EXCITON_NAR_BSE-singlet-TDA-BAR_SCR-full_OC.OUT
    """
    data = numpy_gen_from_txt(name, skip_header=14)
    out = {}
    out["state"] = data[:, 0]
    out["energy"] = data[:, 1]
    out["energy_shifted"] = data[:, 2]
    out["abs_oscillator_strength"] = data[:, 3]
    out["real_oscillator_strength"] = data[:, 4]
    out["imaginary_oscillator_strength"] = data[:, 5]

    return out


def parse_infoxs_out(name: str, parse_timing: bool = False) -> dict:
    """
    Parser for INFOXS.OUT file. Parses only the started and stopped tasks.
    Searches for lines like:
        'EXCITING <version> started for task <taskname> (<tasknumber>)'
    and
        'EXCITING <version> stopped for task <tasknumber>'
    See example file: exciting/test/test_farm/BSE/PBE_SOL-LiF/ref/INFOXS.OUT
    If a started task is found, it gets stored with name, number and status.
    If the task is found to be finished afterwards, the status finished is set to True.

    For success, the last started tasks has to be finished after that (in the file).
    Last finished task is the last task if calculation was successful, the task before that 
    if it finished, else None.
    :param name: path of the file to parse
    :param parse_timing: parse also timing information for the tasks. By default this is set to
                         False. If the task has not finished None is returned as timing.
    :returns: dictionary containing parsed file
    """
    with open(name) as file:
        lines = file.readlines()

    tasks = []
    current_task = -1

    lines = "\n".join(lines)
    all_tasks = re.findall(r'EXCITING .* (started) for task (.*) \( ?(\d+)\)|'
                           r'EXCITING .* stopped for task .* (\d+)', lines)

    for task in all_tasks:
        if task[0] == 'started':
            tasks.append({
                'name': task[1],
                'number': int(task[2]),
                'finished': False
            })
            current_task += 1
        else:
            # asserts shouldn't happen with Exciting:
            assert tasks != [], 'No tasks started!'
            assert tasks[current_task]['number'] == int(task[3]), 'Wrong task stopped.'
            tasks[current_task]['finished'] = True

    success = tasks[-1]['finished']
    last_finished_task = None
    if success:
        last_finished_task = tasks[-1]['name']
    elif len(tasks) > 1 and tasks[-2]['finished']:
        last_finished_task = tasks[-2]['name']

    if parse_timing:
        times = parse_times(lines)
        finished_tasks = [task for task in tasks if task['finished']]
        assert len(times['cpu']) == len(finished_tasks), 'Numbers of finished tasks and parsed times are not the same.'

        for index, task in enumerate(finished_tasks):
            task['cpu_time'] = float(times['cpu'][index])
            task['wall_time'] = float(times['wall'][index])
            task['cpu_time_cum'] = float(times['cpu_cum'][index])
            task['wall_time_cum'] = float(times['wall_cum'][index])
    
    return {'tasks': tasks,
            'success': success,
            'last_finished_task': last_finished_task}


def parse_times(infoxs_string: str) -> dict:
    """Parse the run times in INFOXS.OUT for each task.
    :param infoxs_string: String that contains the INFOXS.OUT file.
    :returns: dictionary containing a list of run times for each measurement.
    """
    cpu_times = re.findall(r'CPU time \s*: ([\d\.\d]+) sec', infoxs_string)
    wall_times = re.findall(r'wall time \s*: ([\d\.\d]+) sec', infoxs_string)
    cpu_times_cum = re.findall(r'CPU time \s* \(cumulative\) \s*: ([\d\.\d]+) sec', infoxs_string)
    wall_times_cum = re.findall(r'wall time \(cumulative\) \s*: ([\d\.\d]+) sec', infoxs_string)

    assert len(cpu_times) == len(wall_times), 'Numbers of parsed timings are not consistent.'
    assert len(cpu_times) == len(cpu_times_cum), 'Numbers of parsed timings are not consistent.'
    assert len(cpu_times) == len(wall_times_cum), 'Numbers of parsed timings are not consistent.'
    
    return {'cpu': cpu_times, 
            'wall': wall_times,
            'cpu_cum': cpu_times_cum,
            'wall_cum': wall_times_cum}



import numpy as np
import pytest
from typing import List

from excitingtools.dataclasses.data_structs import NumberOfStates, PointIndex, BandIndices
from excitingtools.dataclasses.eigenvalues import EigenValues


@pytest.fixture
def eigenvalues_instance():
    """ Initialise an instance of EigenValues, and check attributes are set correctly.
    Reference data taken from `evalqp_oxygen` in test_gw_eigenvalues.py

    G0W0 band structure

    Band index of VBM:  21
    Band index of CBm:  22

    Indirect BandGap (eV):                    5.3790
    at k(VBM) =    0.000   0.500   0.500 ik =     3
       k(CBm) =    0.000   0.000   0.000 ik =     1
    Direct Bandgap at k(VBM) (eV):            5.5451
    Direct Bandgap at k(CBm) (eV):            5.9468
    """
    ref_gw_eigenvalues = np.array([[-0.12905, -0.12891, -0.12896,  0.08958,  0.08957,  0.18396],
                                   [-0.19801, -0.17047, -0.17035,  0.11429,  0.11430,  0.19514],
                                   [-0.15828, -0.15818, -0.10809,  0.09569,  0.14613,  0.18404]])

    ref_k_points = np.array([[0.000000, 0.000000, 0.000000],
                             [0.000000, 0.000000, 0.500000],
                             [0.000000, 0.500000, 0.500000]])

    ref_weights = [0.125, 0.5, 0.375000]

    ref_occupations = np.array([[2, 2, 2, 0, 0, 0], [2, 2, 2, 0, 0, 0], [2, 2, 2, 0, 0, 0]])

    eigen_values = EigenValues(NumberOfStates(19, 24), ref_k_points, [1, 2, 3], ref_gw_eigenvalues, ref_weights,
                               ref_occupations)

    assert ref_gw_eigenvalues.shape == (len(eigen_values.k_points), eigen_values.state_range.n_states)
    assert eigen_values.state_range.first_state == 19
    assert eigen_values.state_range.last_state == 24
    assert np.allclose(eigen_values.k_points, ref_k_points)
    assert eigen_values.k_indices == [1, 2, 3]
    assert np.allclose(eigen_values.all_eigenvalues, ref_gw_eigenvalues), "GW column eigenvalues, for all k-points"
    assert eigen_values.weights == ref_weights
    assert np.allclose(eigen_values.occupations, ref_occupations)
    return eigen_values


def test_class_eigenvalues_get_array_index(eigenvalues_instance):
    assert eigenvalues_instance.get_array_index(19) == 0, "Index of first state in array"
    assert eigenvalues_instance.get_array_index(24) == 5, "Index of last state in array"
    assert eigenvalues_instance.state_range.n_states == 6, "Total number of states"


def test_class_eigenvalues_get_k_point(eigenvalues_instance):
    assert np.allclose(eigenvalues_instance.get_k_point(1), [0.000000, 0.000000, 0.000000])
    assert np.allclose(eigenvalues_instance.get_k_point(2), [0.000000, 0.000000, 0.500000])
    assert np.allclose(eigenvalues_instance.get_k_point(3), [0.000000, 0.500000, 0.500000])


def test_class_eigenvalues_get_index(eigenvalues_instance):
    assert eigenvalues_instance.get_index([0.000000, 0.000000, 0.000000]) == 1
    assert eigenvalues_instance.get_index([0.000000, 0.000000, 0.500000]) == 2
    assert eigenvalues_instance.get_index([0.000000, 0.500000, 0.500000]) == 3
    assert eigenvalues_instance.get_index([0.500000, 0.500000, 0.500000]) == EigenValues.NO_MATCH, "No k-point matched"


def test_class_eigenvalues_get_k_points(eigenvalues_instance):
    k_points_and_indices: List[PointIndex] = eigenvalues_instance.get_k_points()
    assert np.allclose(k_points_and_indices[0].point, [0.000000, 0.000000, 0.000000])
    assert np.allclose(k_points_and_indices[1].point, [0.000000, 0.000000, 0.500000])
    assert np.allclose(k_points_and_indices[2].point, [0.000000, 0.500000, 0.500000])
    assert k_points_and_indices[0].index == 1
    assert k_points_and_indices[1].index == 2
    assert k_points_and_indices[2].index == 3


def test_class_eigenvalues_get_eigenvalues(eigenvalues_instance):
    eigenvalues = eigenvalues_instance.get_eigenvalues(k_point=[0.0, 0.5, 0.5])
    assert np.allclose(eigenvalues,
                       [-0.15828, -0.15818, -0.10809, 0.09569, 0.14613, 0.18404])

    assert eigenvalues_instance.get_eigenvalues(k_point=[0.5, 0.5, 0.5]).size == 0, \
        "No k-point, hence eigenvalues"


def test_class_eigenvalues_band_gap(eigenvalues_instance):
    k_valence = [0.000, 0.500, 0.500]
    k_conduction = [0.000, 0.000, 0.000]

    indirect_band_gap = eigenvalues_instance.band_gap(BandIndices(21, 22), k_points=[k_valence, k_conduction])
    direct_band_gap_at_Gamma = eigenvalues_instance.band_gap(BandIndices(21, 22), k_points=[k_conduction, k_conduction])

    assert np.isclose(indirect_band_gap, 0.19767), 'Indirect band gap in Ha'
    assert np.isclose(direct_band_gap_at_Gamma, 0.218540887), 'Direct band gap at Gamma, in Ha'


def test_class_eigenvalues_get_transition_energy(eigenvalues_instance):
    k_valence = [0.000, 0.500, 0.500]
    k_conduction = [0.000, 0.000, 0.000]

    indirect_gap = eigenvalues_instance.get_transition_energy(k_valence, k_conduction)
    direct_gap_at_Gamma =eigenvalues_instance.get_transition_energy(k_conduction, k_conduction)

    assert np.isclose(indirect_gap, 0.19767), 'Transition energy for X→Γ, in Ha'
    assert np.isclose(direct_gap_at_Gamma, 0.21854), 'Transition energy for Γ→Γ, in Ha'

    with pytest.raises(ValueError, match="Requested conduction k-point \\[1, 1, 1\\] not present."):
        eigenvalues_instance.get_transition_energy(k_valence, [1, 1, 1]), \
        "ValueError is returned if k-point is not matched."



""" Main exciting input classes. """
from excitingtools.input.input_classes import (
    ExcitingGroundStateInput,
    ExcitingXSInput,
    ExcitingPropertiesInput,
    ExcitingRelaxInput,
    ExcitingPhononsInput,
    ExcitingGWInput,
    ExcitingMDInput,
    ExcitingEPHInput
)
from excitingtools.input.input_xml import ExcitingInputXML
from excitingtools.input.structure import ExcitingStructure



""" Math functions.
"""
import numpy as np


def triple_product(a, b, c) -> np.ndarray:
    r"""
    Vector triple product, defined as 
      \mathbf{a} \cdot (\mathbf{b} \wedge \mathbf{c})

    :param a: Vector a 
    :param b: Vector b
    :param c: Vector c
    :return triple product
    """
    return np.dot(a, np.cross(b, c))


def unit_vector(x: np.ndarray) -> np.ndarray:
    r"""
    Unit vector of a vector 'x' 
      \mathbf{\hat{x}} =  \frac{\mathbf{x}}{|\mathbf{x}|}

    :param x: Vector x
    :result: Unit vector of x
    """
    return x / np.linalg.norm(x)



""" Test properties.

NOTE:
All attribute tests should assert on the XML tree content's as the attribute
order is not preserved by the ElementTree.tostring method. Elements appear to
be fine.
"""
import re

import pytest

from excitingtools.exciting_dict_parsers.input_parser import parse_element_xml
from excitingtools.input.bandstructure import band_structure_input_from_ase_atoms_obj, \
    get_bandstructure_input_from_exciting_structure, band_structure_input_from_cell_or_bandpath
from excitingtools.input.input_classes import ExcitingPropertiesInput, ExcitingGroundStateInput
from excitingtools.input.input_xml import ExcitingInputXML
from excitingtools.input.structure import ExcitingStructure


@pytest.fixture
def ase_ag():
    """If we cannot import ase we skip tests with this fixture.

    :returns: bulk Silver crystal
    """
    ase_build = pytest.importorskip('ase.build')
    return ase_build.bulk('Ag')


def test_bandstructure_properties_input():
    """Test giving the bandstructure input as a nested dict."""
    properties = {
        "bandstructure": {"plot1d": {"path": {
            "steps": 100,
            "point":
                [
                    {"coord": [1, 0, 0], "label": "Gamma"},
                    {"coord": [0.625, 0.375, 0], "label": "K"},
                    {"coord": [0.5, 0.5, 0], "label": "X", "breakafter": True},
                    {"coord": [0, 0, 0], "label": "Gamma"},
                    {"coord": [0.5, 0, 0], "label": "L"}]}}}}

    properties_input = ExcitingPropertiesInput(**properties)
    xml_string = properties_input.to_xml_str()

    assert properties == parse_element_xml(xml_string)


def test_bs_from_ase(ase_ag):
    """Test getting XML bandstructure input for elemental silver.

    We test first that the bandstructure path defined by the high symmetry
    points is correct. We also check the high symmetry point names are
    correct.
    """
    ase_ag.set_cell(ase_ag.cell * 3)
    bs = band_structure_input_from_ase_atoms_obj(ase_ag)
    properties_input = ExcitingPropertiesInput(bandstructure=bs)
    bs_xml_string = properties_input.to_xml_str()
    # Get the high symmetry path.
    all_labels = re.findall(r'label=\"([A-Z])', bs_xml_string)
    assert all_labels == ['G', 'X', 'W', 'K', 'G', 'L', 'U', 'W', 'L', 'K', 'U', 'X']
    # Ensure there is only one breakafter statement.
    assert len(re.findall(r'breakafter', bs_xml_string)) == 1

    bs_xml = properties_input.to_xml()
    points = bs_xml.find("bandstructure").find("plot1d").find("path").findall("point")
    assert len(points) == 12
    assert points[0].get("coord") == "0.0 0.0 0.0"
    assert points[1].get("coord") == "0.5 0.0 0.5"
    assert points[2].get("coord") == "0.5 0.25 0.75"
    assert points[3].get("coord") == "0.375 0.375 0.75"
    assert points[4].get("coord") == "0.0 0.0 0.0"
    assert points[5].get("coord") == "0.5 0.5 0.5"
    assert points[6].get("coord") == "0.625 0.25 0.625"
    assert points[7].get("coord") == "0.5 0.25 0.75"
    assert points[8].get("coord") == "0.5 0.5 0.5"
    assert points[9].get("coord") == "0.375 0.375 0.75"
    assert points[9].get("breakafter") == "true"
    assert points[10].get("coord") == "0.625 0.25 0.625"
    assert points[11].get("coord") == "0.5 0.0 0.5"


def test_bs_input(ase_ag):
    """Test writing exciting full input xml file with bandstructure property."""
    gs = ExcitingGroundStateInput(rgkmax=5.0)
    struct = ExcitingStructure(ase_ag)
    bs = band_structure_input_from_ase_atoms_obj(ase_ag)
    properties_input = ExcitingPropertiesInput(bandstructure=bs)
    input_xml = ExcitingInputXML(
        structure=struct, groundstate=gs, title="BS exciting",
        properties=properties_input).to_xml()

    assert input_xml.tag == 'input'
    assert input_xml.keys() == []

    subelements = list(input_xml)
    assert len(subelements) == 4
    title_xml = subelements[0]
    assert title_xml.tag == 'title'
    assert title_xml.keys() == []
    assert title_xml.text == 'BS exciting'

    properties_xml = subelements[3]
    assert properties_xml.tag == 'properties'
    assert properties_xml.keys() == []
    assert properties_xml.text == ' '
    assert properties_xml[0].tag == 'bandstructure'
    assert properties_xml[0][0].tag == 'plot1d'
    assert properties_xml[0][0][0].tag == 'path'

    first_coordinates = properties_xml[0][0][0][0].get('coord')
    first_label = properties_xml[0][0][0][0].get('label')
    assert first_coordinates == '0.0 0.0 0.0'
    assert first_label == 'G'


def test_get_bandstructure_input_from_exciting_structure(ase_ag):
    structure = ExcitingStructure(ase_ag)
    bs_xml = get_bandstructure_input_from_exciting_structure(structure).to_xml()
    points = bs_xml.find("plot1d").find("path").findall("point")
    assert len(points) == 12
    assert points[0].get("coord") == "0.0 0.0 0.0"
    assert points[1].get("coord") == "0.5 0.0 0.5"


def test_get_bandstructure_input_from_exciting_structure_stretch(ase_ag):
    structure = ExcitingStructure(ase_ag)
    structure.crystal_properties.stretch = [2, 1, 1]
    bs_xml = get_bandstructure_input_from_exciting_structure(structure).to_xml()
    points = bs_xml.find("plot1d").find("path").findall("point")
    assert len(points) == 13
    assert points[0].get("coord") == "0.0 0.0 0.0"
    assert points[1].get("coord") == "0.5 0.0 0.5"
    assert points[2].get("coord") == "0.125 -0.375 0.3125"


def test_get_bandstructure_input_from_ase_bandpath(ase_ag):
    """Test creating required bandstructure input from an ASE bandpath.
    
    Someone could create a bandpath that doesn't fall on special points
    e.g., 'Kpt0Kpt1,GMX,XZ'.
    """
    ase = pytest.importorskip("ase")   
    bandpath = ase.dft.kpoints.bandpath(
        cell=ase_ag.cell,
        path=[(0, 0, 0), (0.1, 0.2, 0.1), (0, 0, 0)])
    bandstructure = band_structure_input_from_cell_or_bandpath(bandpath)
    bs_xml = bandstructure.to_xml()
    assert bs_xml.tag == "bandstructure", 'Root tag should be "bandstructure"'
    plot1d_xml = bs_xml.find("plot1d")
    assert plot1d_xml is not None, 'Missing "plot1d" subtree'

    path_xml = plot1d_xml.find("path")
    assert path_xml is not None, 'Missing "path" subtree'
    assert path_xml.get("steps") == "100", 'Invalid value for "steps" attribute'

    assert len(list(path_xml)) == 3
    point1 = path_xml[0]
    assert point1.get("coord") == "0.0 0.0 0.0", 'Invalid value for "coord" attribute of point 1'
    assert point1.get("label") == "G", 'Invalid value for "label" attribute of point 1'

    point2 = path_xml[1]
    assert point2.get("coord") == "0.1 0.2 0.1", 'Invalid value for "coord" attribute of point 2'
    assert point2.get("label") == "Kpt0", 'Invalid value for "label" attribute of point 2'

    point3 = path_xml[2]
    assert point3.get("coord") == "0.0 0.0 0.0", 'Invalid value for "coord" attribute of point 3'
    assert point3.get("label") == "G", 'Invalid value for "label" attribute of point 3'



""" Binary runner and results classes.
"""
from __future__ import annotations

import copy
import enum
import os
import shutil
import subprocess
import time
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union

from excitingtools.utils.jobflow_utils import special_serialization_attrs


class RunnerCode(enum.Enum):
    """ Runner codes.
     By default, the initial value starts at 1.
    """
    time_out = enum.auto()


@dataclass
class SubprocessRunResults:
    """ Results returned from subprocess.run()
    """

    stdout: str | bytes
    stderr: str | bytes
    return_code: int | RunnerCode
    process_time: float = None

    def __post_init__(self):
        self.success = self.return_code == 0


class BinaryRunner:
    """ Class to execute a subprocess.
    """
    path_type = Union[str, Path]

    def __init__(self,
                 binary: path_type,
                 run_cmd: List[str] | str = "",
                 omp_num_threads: int = 1,
                 time_out: int = 60,
                 directory: path_type = './',
                 args: Optional[List[str]] = None):
        """ Initialise class.

        :param str binary: Binary name prepended by full path, or just binary name (if present in $PATH).
        :param Union[List[str], str] run_cmd: Run commands sequentially as a list. For example:
          * For serial: []
          * For MPI:   ['mpirun', '-np', '2']
        or as a string. For example"
          * For serial: ""
          * For MPI: "mpirun -np 2"
        :param omp_num_threads: Number of OMP threads.
        :param time_out: Number of seconds before a job is defined to have timed out.
        :param args: Optional arguments for the binary.
        """
        if args is None:
            args = []
        self.binary = Path(binary).as_posix()
        self.directory = directory
        self.run_cmd = run_cmd
        self.omp_num_threads = omp_num_threads
        self.time_out = time_out
        self.args = args

        if not os.path.isfile(self.binary):
            # If just the binary name, try checking the $PATH
            self.binary = shutil.which(self.binary)
            if not self.binary:
                raise FileNotFoundError(
                    f"{binary} binary is not present in the current directory nor in $PATH"
                )

        if not Path(directory).is_dir():
            raise OSError(f"Run directory does not exist: {directory}")

        if isinstance(run_cmd, str):
            self.run_cmd = run_cmd.split()
        elif not isinstance(run_cmd, list):
            raise ValueError(
                "Run commands expected in a str or list. For example ['mpirun', '-np', '2']"
            )

        self._check_mpi_processes()

        if omp_num_threads <= 0:
            raise ValueError("Number of OMP threads must be > 0")

        if time_out <= 0:
            raise ValueError("time_out must be a positive integer")

    def as_dict(self) -> dict:
        """Returns a dictionary representing the current object for later recreation.
        The serialise attributes are required for recognition by monty and jobflow.
        """
        serialise_attrs = special_serialization_attrs(self)
        return {**serialise_attrs, **self.__dict__}

    @classmethod
    def from_dict(cls, d: dict):
        my_dict = copy.deepcopy(d)
        # Remove key value pairs needed for workflow programs
        # call function on class to get only the keys (values not needed)
        serialise_keys = special_serialization_attrs(cls)
        for key in serialise_keys:
            my_dict.pop(key, None)
        return cls(**my_dict)

    def _check_mpi_processes(self):
        """ Check whether mpi is specified and if yes that the number of MPI processes specified is valid.
        """
        # Search if MPI is specified:
        try:
            i = self.run_cmd.index('-np')
        except ValueError:
            # .index will return ValueError if 'np' not found. This corresponds to serial and omp calculations.
            return
        try:
            mpi_processes = int(self.run_cmd[i + 1])
        except IndexError:
            raise ValueError("Number of MPI processes must be specified after the '-np'")
        except ValueError:
            raise ValueError("Number of MPI processes should be an int")
        if mpi_processes <= 0:
            raise ValueError("Number of MPI processes must be > 0")

    def run(self) -> SubprocessRunResults:
        """Run a binary.
        """
        execution_list = self.run_cmd + [self.binary] + self.args
        my_env = {**os.environ, "OMP_NUM_THREADS": str(self.omp_num_threads)}

        time_start: float = time.time()
        try:
            result = subprocess.run(execution_list,
                                    env=my_env,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    timeout=self.time_out,
                                    cwd=self.directory)
            total_time = time.time() - time_start
            return SubprocessRunResults(result.stdout, result.stderr,
                                        result.returncode, total_time)

        except subprocess.TimeoutExpired as timed_out:
            error = 'BinaryRunner: Job timed out. \n\n'
            if timed_out.stderr:
                error += timed_out.stderr.decode("utf-8")
            return SubprocessRunResults(timed_out.output, error,
                                        RunnerCode.time_out, self.time_out)



# from excitingtools.exciting_obj_parsers.gw_eigenvalues import gw_eigenvalue_parser, OxygenEvalQPColumns, \
# NitrogenEvalQPColumns
from excitingtools.exciting_obj_parsers.ks_band_structure import parse_band_structure
